import argparse
import os
from torch.cuda import device_count
import torch.multiprocessing as mp
from konfai.utils.utils import setup, TensorBoard, Log
from konfai import KONFAI_NB_CORES

import sys
sys.path.insert(0, os.getcwd())

def main():
    parser = argparse.ArgumentParser(description="KonfAI", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    try:
        with setup(parser) as distributedObject:
            with Log(distributedObject.name, 0):
                world_size = device_count()
                if world_size == 0:
                    world_size = int(KONFAI_NB_CORES())
                distributedObject.setup(world_size)
                with TensorBoard(distributedObject.name):
                    mp.spawn(distributedObject, nprocs=world_size)
    except KeyboardInterrupt:
        print("\n[KonfAI] Manual interruption (Ctrl+C)")

def cluster():
    parser = argparse.ArgumentParser(description="KonfAI", formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # Cluster manager arguments
    cluster_args = parser.add_argument_group('Cluster manager arguments')
    cluster_args.add_argument('--name', type=str, help='Task name', required=True)
    cluster_args.add_argument('--num-nodes', '--num_nodes', default=1, type=int, help='Number of nodes')
    cluster_args.add_argument('--memory', type=int, default=16, help='Amount of memory per node')
    cluster_args.add_argument('--time-limit', '--time_limit', type=int, default=1440, help='Job time limit in minute')
    cluster_args.add_argument('--resubmit', action='store_true', help='Automatically resubmit job just before timout')
    try:
        with setup(parser) as distributedObject:
            args = parser.parse_args()
            config = vars(args)
            os.environ["KONFAI_OVERWRITE"] = "True"
            os.environ["KONFAI_CLUSTER"] = "True"

            n_gpu = len(config["gpu"].split(","))
            distributedObject.setup(n_gpu*int(config["num_nodes"]))
            import submitit
            executor = submitit.AutoExecutor(folder="./Cluster/")
            executor.update_parameters(name=config["name"], mem_gb=config["memory"], gpus_per_node=n_gpu, tasks_per_node=n_gpu//distributedObject.size, cpus_per_task=config["num_workers"], nodes=config["num_nodes"], timeout_min=config["time_limit"])
            with TensorBoard(distributedObject.name):
                executor.submit(distributedObject)
    except KeyboardInterrupt:
        print("\n[KonfAI] Manual interruption (Ctrl+C)")
    except Exception as e:
        print(e)
        exit(1)


import SimpleITK as sitk
from typing import Union
import numpy as np
import sys
import scipy

def parameterMap_to_transform(path_src: str) -> Union[sitk.Transform, list[sitk.Transform]]:
    transform = sitk.ReadParameterFile(path_src)
    format = lambda x: [float(i) for i in x]
    dimension = int(transform["FixedImageDimension"][0])

    if transform["Transform"][0] == "EulerTransform":
        if dimension == 2:
            result = sitk.Euler2DTransform()
        else:
            result = sitk.Euler3DTransform()
        parameters = format(transform["TransformParameters"])
        fixedParameters = format(transform["CenterOfRotationPoint"])+[0]
    elif transform["Transform"][0] == "TranslationTransform":
        result = sitk.TranslationTransform(dimension)
        parameters = format(transform["TransformParameters"])
        fixedParameters = []
    elif transform["Transform"][0] == "AffineTransform":
        result = sitk.AffineTransform(dimension)
        parameters = format(transform["TransformParameters"])
        fixedParameters = format(transform["CenterOfRotationPoint"])+[0]
    elif transform["Transform"][0] == "BSplineStackTransform":
        parameters = format(transform["TransformParameters"])
        GridSize = format(transform["GridSize"])
        GridOrigin = format(transform["GridOrigin"])
        GridSpacing = format(transform["GridSpacing"])
        GridDirection =  np.asarray(format(transform["GridDirection"])).reshape((dimension, dimension)).T.flatten() 
        fixedParameters = np.concatenate([GridSize, GridOrigin, GridSpacing, GridDirection])

        nb = int(format(transform["Size"])[-1])
        sub = int(np.prod(GridSize))*dimension
        results = []
        for i in range(nb):
            result = sitk.BSplineTransform(dimension)
            sub_parameters = np.asarray(parameters[i*sub:(i+1)*sub])
            result.SetFixedParameters(fixedParameters)
            result.SetParameters(sub_parameters)
            results.append(result)
        return results
    elif transform["Transform"][0] == "AffineLogStackTransform":
        parameters = format(transform["TransformParameters"])
        fixedParameters = format(transform["CenterOfRotationPoint"])+[0]

        nb = int(transform["NumberOfSubTransforms"][0])
        sub = dimension*4
        results = []
        for i in range(nb):
            result = sitk.AffineTransform(dimension)
            sub_parameters = np.asarray(parameters[i*sub:(i+1)*sub])

            result.SetFixedParameters(fixedParameters)
            result.SetParameters(np.concatenate([scipy.linalg.expm(sub_parameters[:dimension*dimension].reshape((dimension,dimension))).flatten(), sub_parameters[-dimension:]]))
            results.append(result)
        return results
    else:
        result = sitk.BSplineTransform(dimension)
        
        parameters = format(transform["TransformParameters"])
        GridSize = format(transform["GridSize"])
        GridOrigin = format(transform["GridOrigin"])
        GridSpacing = format(transform["GridSpacing"])
        GridDirection = np.array(format(transform["GridDirection"])).reshape((dimension,dimension)).T.flatten() 
        fixedParameters = np.concatenate([GridSize, GridOrigin, GridSpacing, GridDirection])

    result.SetFixedParameters(fixedParameters)
    result.SetParameters(parameters)
    return result

if __name__ == "__main__":
    out_path = sys.argv[1]
    finename = sys.argv[2]
    finename_dest = sys.argv[3]
    transform = parameterMap_to_transform("{}/{}".format(out_path, finename))
    sitk.WriteTransform(transform, "{}/{}".format(out_path, finename_dest))

def getFlatLabel(mask: sitk.Image, labels: list[int]) -> sitk.Image:
    data = sitk.GetArrayFromImage(mask)
    result_data = np.zeros_like(data, np.uint8)

    for label in labels:
        result_data[data == label] = 1

    result = sitk.GetImageFromArray(result_data)
    result.CopyInformation(mask)        
    return result

def rampFilterHistogram(image: sitk.Image, rampStart: float, rampEnd: float) -> sitk.Image:
    imageData = sitk.GetArrayFromImage(image)
    filter = np.logical_and(imageData > rampStart, imageData < rampEnd)
    rampWidth = rampEnd - rampStart
    imageData[filter] = (1/rampWidth) * (imageData[filter] - rampStart) * imageData[filter]
    filteredImage = sitk.GetImageFromArray(imageData)
    filteredImage.CopyInformation(image)
    return filteredImage

def elastic_registration(   fixed_image : sitk.Image,
                            moving_image : sitk.Image,
                            fixed_mask : Union[sitk.Image, None],
                            moving_mask : Union[sitk.Image, None],
                            name_parameterMap : str,
                            outputDir: str) -> sitk.Transform:
        labels = np.unique(sitk.GetArrayFromImage(fixed_mask))
        fixed_mask = getFlatLabel(fixed_mask, labels[1:])
        moving_mask = getFlatLabel(moving_mask, labels[1:])

        fixed_mask.CopyInformation(fixed_image)
        moving_mask.CopyInformation(moving_image)

        fixed_mask_dillated = sitk.BinaryDilate(fixed_mask, [5,5,5])
        moving_mask_dillated = sitk.BinaryDilate(moving_mask, [5,5,5])
        
        fixed_image = sitk.Mask(fixed_image, fixed_mask)
        moving_image = sitk.Mask(moving_image, moving_mask)


        minGradientMagnitude = 50
        fixed_image_gradient = rampFilterHistogram(sitk.VectorMagnitude(sitk.Gradient(fixed_image)), 0, minGradientMagnitude)
        moving_image_gradient = rampFilterHistogram(sitk.VectorMagnitude(sitk.Gradient(moving_image)), 0, minGradientMagnitude)

        elastixImageFilter = sitk.ElastixImageFilter()
        elastixImageFilter.SetFixedImage(fixed_image_gradient)
        elastixImageFilter.AddFixedImage(fixed_image)
        elastixImageFilter.AddFixedImage(fixed_image)

        if fixed_mask is not None:
            elastixImageFilter.SetFixedMask(fixed_mask)
            elastixImageFilter.AddFixedMask(fixed_mask_dillated)
            elastixImageFilter.AddFixedMask(fixed_mask_dillated)


        elastixImageFilter.SetMovingImage(moving_image_gradient)
        elastixImageFilter.AddMovingImage(moving_image)
        elastixImageFilter.AddMovingImage(moving_image)
        
        if moving_mask is not None:
            elastixImageFilter.SetMovingMask(moving_mask)
            elastixImageFilter.AddMovingMask(moving_mask_dillated)
            elastixImageFilter.AddMovingMask(moving_mask_dillated)

        elastixImageFilter.SetParameterMap(sitk.ReadParameterFile("{}.txt".format(name_parameterMap)))
        elastixImageFilter.LogToConsoleOn()
        elastixImageFilter.SetOutputDirectory(outputDir)
        
        elastixImageFilter.Execute()
        
        transform = parameterMap_to_transform("{}TransformParameters".format(outputDir))
        
        return transform

def registration(   fixed_image : sitk.Image,
                    moving_image : sitk.Image,
                    fixed_mask : Union[sitk.Image, None],
                    moving_mask : Union[sitk.Image, None],
                    name_parameterMap : str,
                    outputDir: str) -> sitk.Transform:
        elastixImageFilter = sitk.ElastixImageFilter()
        elastixImageFilter.SetFixedImage(fixed_image)
        if fixed_mask is not None:
            elastixImageFilter.SetFixedMask(fixed_mask)

        elastixImageFilter.SetMovingImage(moving_image)
        if moving_mask is not None:
            elastixImageFilter.SetMovingMask(moving_mask)

        elastixImageFilter.SetParameterMap(sitk.ReadParameterFile("{}.txt".format(name_parameterMap)))
        elastixImageFilter.LogToConsoleOn()
        elastixImageFilter.SetOutputDirectory(outputDir)
        elastixImageFilter.Execute()
        
        transform = parameterMap_to_transform("{}TransformParameters".format(outputDir))
        
        return transform

def registration_groupewise(images_1: sitk.Image, masks: sitk.Image, images_2: sitk.Image, name_parameterMap : str, output_dir: str):
    elastixImageFilter = sitk.ElastixImageFilter()
    elastixImageFilter.SetFixedImage(images_1)
    elastixImageFilter.SetMovingImage(images_1)
    elastixImageFilter.SetFixedMask(masks)
    elastixImageFilter.SetMovingMask(masks)
    
    if images_2 is not None:
        elastixImageFilter.AddFixedImage(images_2)
        elastixImageFilter.AddMovingImage(images_2)
        #elastixImageFilter.AddFixedImage(images_2)
        #elastixImageFilter.AddMovingImage(images_2)
    
    elastixImageFilter.SetParameterMap(sitk.ReadParameterFile("{}.txt".format(name_parameterMap)))
    elastixImageFilter.LogToConsoleOn()
    elastixImageFilter.SetOutputDirectory(output_dir)
    elastixImageFilter.LogToFileOn()
    elastixImageFilter.Execute()

    transforms = parameterMap_to_transform("{}TransformParameters".format(output_dir))
    return transforms


from abc import ABC, abstractmethod
import builtins
import importlib
import shutil
import torch
import tqdm
import os

from konfai import MODELS_DIRECTORY, PREDICTIONS_DIRECTORY, CONFIG_FILE, MODEL, KONFAI_ROOT
from konfai.utils.config import config
from konfai.utils.utils import State, get_patch_slices_from_nb_patch_per_dim, NeedDevice, _getModule, DistributedObject, DataLog, description, PredictorError
from konfai.utils.dataset import Dataset, Attribute
from konfai.data.data_manager import DataPrediction, DatasetIter
from konfai.data.patching import Accumulator, PathCombine
from konfai.network.network import ModelLoader, Network, NetState, CPU_Model
from konfai.data.transform import Transform, TransformLoader

from torch.utils.tensorboard.writer import SummaryWriter
from typing import Union
import numpy as np
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel as DDP
from torch.utils.data import DataLoader
import importlib
import copy
from collections import defaultdict

class OutDataset(Dataset, NeedDevice, ABC):

    def __init__(self, filename: str, group: str, before_reduction_transforms : dict[str, TransformLoader], after_reduction_transforms : dict[str, TransformLoader], final_transforms : dict[str, TransformLoader], patchCombine: Union[str, None], reduction: str) -> None: 
        filename, format = filename.split(":")
        super().__init__(filename, format)
        self.group = group
        self._before_reduction_transforms = before_reduction_transforms
        self._after_reduction_transforms = after_reduction_transforms
        self._final_transforms = final_transforms
        self._patchCombine = patchCombine
        self.reduction_classpath = reduction
        self.reduction = None
       
        self.before_reduction_transforms : list[Transform] = []
        self.after_reduction_transforms : list[Transform] = []
        self.final_transforms : list[Transform] = []
        self.patchCombine: PathCombine = None

        self.output_layer_accumulator: dict[int, dict[int, Accumulator]] = {}
        self.attributes: dict[int, dict[int, dict[int, Attribute]]] = {}
        self.names: dict[int, str] = {}
        self.nb_data_augmentation = 0

    def load(self, name_layer: str, datasets: list[Dataset], groups: dict[str, str]):
        transforms_type = ["before_reduction_transforms", "after_reduction_transforms", "final_transforms"]
        for name, _transform_type, transform_type in [(k, getattr(self, "_{}".format(k)), getattr(self, k)) for k in transforms_type]:
            
            if _transform_type is not None:
                for classpath, transform in _transform_type.items():
                    transform = transform.getTransform(classpath, DL_args =  "{}.outsDataset.{}.OutDataset.{}".format(KONFAI_ROOT(), name_layer, name))
                    transform.setDatasets(datasets)
                    transform_type.append(transform)

        if self._patchCombine is not None:
            module, name = _getModule(self._patchCombine, "konfai.data.patching")                                                      
            self.patchCombine = config("{}.outsDataset.{}.OutDataset".format(KONFAI_ROOT(), name_layer))(getattr(importlib.import_module(module), name))(config = None)

        module, name = _getModule(self.reduction_classpath, "konfai.predictor")
        if module == "konfai.predictor":
            self.reduction = getattr(importlib.import_module(module), name)
        else:
            self.reduction = config("{}.outsDataset.{}.OutDataset.{}".format(KONFAI_ROOT(), name_layer, self.reduction_classpath))(getattr(importlib.import_module(module), name))(config = None)
       
    
    def setPatchConfig(self, patchSize: Union[list[int], None], overlap: Union[int, None], nb_data_augmentation: int) -> None:
        if patchSize is not None and overlap is not None:
            if self.patchCombine is not None:
                self.patchCombine.setPatchConfig(patchSize, overlap)
        else:
            self.patchCombine = None
        self.nb_data_augmentation = nb_data_augmentation
    
    def setDevice(self, device: torch.device):
        super().setDevice(device)
        transforms_type = ["before_reduction_transforms", "after_reduction_transforms", "final_transforms"]
        for transform_type in [(getattr(self, k)) for k in transforms_type]:
            if transform_type is not None:
                for transform in transform_type:
                    transform.setDevice(device)       

    @abstractmethod
    def addLayer(self, index_dataset: int, index_augmentation: int, index_patch: int, layer: torch.Tensor, dataset: DatasetIter):
        pass

    def isDone(self, index: int) -> bool:
        return len(self.output_layer_accumulator[index]) == self.nb_data_augmentation and all([acc.isFull() for acc in self.output_layer_accumulator[index].values()])

    @abstractmethod
    def getOutput(self, index: int, dataset: DatasetIter) -> torch.Tensor:
        pass

    def write(self, index: int, name: str, layer: torch.Tensor):
        super().write(self.group, name, layer.numpy(), self.attributes[index][0][0])
        self.attributes.pop(index)

class Reduction():

    def __init__(self):
        pass

class Mean(Reduction):

    def __init__(self):
        pass

    def __call__(self, result: torch.Tensor) -> torch.Tensor:
        return torch.mean(result.float(), dim=0)
        
class Median(Reduction):

    def __init__(self):
        pass
    
    def __call__(self, result: torch.Tensor) -> torch.Tensor:
        return torch.median(result.float(), dim=0).values

class OutSameAsGroupDataset(OutDataset):

    @config("OutDataset")
    def __init__(self, sameAsGroup: str = "default", dataset_filename: str = "default:./Dataset:mha", group: str = "default", before_reduction_transforms : dict[str, TransformLoader] = {"default:Normalize": TransformLoader()}, after_reduction_transforms : dict[str, TransformLoader] = {"default:Normalize": TransformLoader()}, final_transforms : dict[str, TransformLoader] = {"default:Normalize": TransformLoader()}, patchCombine: Union[str, None] = None, reduction: str = "mean", inverse_transform: bool = True) -> None:
        super().__init__(dataset_filename, group, before_reduction_transforms, after_reduction_transforms, final_transforms, patchCombine, reduction)
        self.group_src, self.group_dest = sameAsGroup.split(":")
        self.inverse_transform = inverse_transform

    def addLayer(self, index_dataset: int, index_augmentation: int, index_patch: int, layer: torch.Tensor, dataset: DatasetIter):
        if index_dataset not in self.output_layer_accumulator or index_augmentation not in self.output_layer_accumulator[index_dataset]:
            input_dataset = dataset.getDatasetFromIndex(self.group_dest, index_dataset)
            if index_dataset not in self.output_layer_accumulator:
                self.output_layer_accumulator[index_dataset] = {}
                self.attributes[index_dataset] = {}
                self.names[index_dataset] = input_dataset.name
            self.attributes[index_dataset][index_augmentation] = {}

            self.output_layer_accumulator[index_dataset][index_augmentation] = Accumulator(input_dataset.patch.getPatch_slices(index_augmentation), input_dataset.patch.patch_size, self.patchCombine, batch=False)

            for i in range(len(input_dataset.patch.getPatch_slices(index_augmentation))):
                self.attributes[index_dataset][index_augmentation][i] = Attribute(input_dataset.cache_attributes[0])

        if self.inverse_transform:
            for transform in reversed(dataset.groups_src[self.group_src][self.group_dest].post_transforms):
                layer = transform.inverse(self.names[index_dataset], layer, self.attributes[index_dataset][index_augmentation][index_patch])
        self.output_layer_accumulator[index_dataset][index_augmentation].addLayer(index_patch, layer)


    def load(self, name_layer: str, datasets: list[Dataset], groups: dict[str, str]):
        super().load(name_layer, datasets, groups)
        
        if self.group_src not in groups.keys():
            raise PredictorError(
                f"Source group '{self.group_src}' not found. Available groups: {list(groups.keys())}."
            )

        if self.group_dest not in groups[self.group_src]:
            raise PredictorError(
                f"Destination group '{self.group_dest}' not found. Available groups: {groups[self.group_src]}."
            )
    
    def _getOutput(self, index: int, index_augmentation: int, dataset: DatasetIter) -> torch.Tensor:
        layer = self.output_layer_accumulator[index][index_augmentation].assemble()
        if index_augmentation > 0:
            
            i = 0
            index_augmentation_tmp = index_augmentation-1
            for dataAugmentations in dataset.dataAugmentationsList:
                if index_augmentation_tmp >= i and index_augmentation_tmp < i+dataAugmentations.nb:
                    for dataAugmentation in reversed(dataAugmentations.dataAugmentations):
                        layer = dataAugmentation.inverse(index, index_augmentation_tmp-i, layer)
                    break
                i += dataAugmentations.nb

        for transform in self.before_reduction_transforms:
            layer = transform(self.names[index], layer, self.attributes[index][index_augmentation][0])
        
        return layer

    def getOutput(self, index: int, dataset: DatasetIter) -> torch.Tensor:
        result = torch.cat([self._getOutput(index, index_augmentation, dataset).unsqueeze(0) for index_augmentation in self.output_layer_accumulator[index].keys()], dim=0)
        self.output_layer_accumulator.pop(index)
        result = self.reduction(result.float()).to(result.dtype)
            
        for transform in self.after_reduction_transforms:
            result = transform(self.names[index], result, self.attributes[index][0][0])

        if self.inverse_transform:
            for transform in reversed(dataset.groups_src[self.group_src][self.group_dest].pre_transforms):
                result = transform.inverse(self.names[index], result, self.attributes[index][0][0])
        
        for transform in self.final_transforms:
            result = transform(self.names[index], result, self.attributes[index][0][0]) 
        return result

class OutDatasetLoader():

    @config("OutDataset")
    def __init__(self, name_class: str = "OutSameAsGroupDataset") -> None:
        self.name_class = name_class

    def getOutDataset(self, layer_name: str) -> OutDataset:
        return getattr(importlib.import_module("konfai.predictor"), self.name_class)(config = None, DL_args = "Predictor.outsDataset.{}".format(layer_name))

class _Predictor():

    def __init__(self, world_size: int, global_rank: int, local_rank: int, autocast: bool, predict_path: str, data_log: Union[list[str], None], outsDataset: dict[str, OutDataset], modelComposite: DDP, dataloader_prediction: DataLoader) -> None:
        self.world_size = world_size        
        self.global_rank = global_rank
        self.local_rank = local_rank

        self.modelComposite = modelComposite
        self.dataloader_prediction = dataloader_prediction
        self.outsDataset = outsDataset
        self.autocast = autocast
        
        self.it = 0

        self.device = self.modelComposite.device
        self.dataset: DatasetIter = self.dataloader_prediction.dataset
        patch_size, overlap = self.dataset.getPatchConfig()
        for outDataset in self.outsDataset.values():
            outDataset.setPatchConfig([size for size in patch_size if size > 1], overlap, np.max([int(np.sum([data_augmentation.nb for data_augmentation in self.dataset.dataAugmentationsList])+1), 1]))
        self.data_log : dict[str, tuple[DataLog, int]] = {}
        if data_log is not None:
            for data in data_log:
                self.data_log[data.split("/")[0].replace(":", ".")] = (DataLog.__getitem__(data.split("/")[1]).value[0], int(data.split("/")[2]))
        self.tb = SummaryWriter(log_dir = predict_path+"Metric/") if len([network for network in self.modelComposite.module.getNetworks().values() if network.measure is not None]) or len(self.data_log) else None
        
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        if self.tb:
            self.tb.close()

    def getInput(self, data_dict : dict[str, tuple[torch.Tensor, int, int, int, str, bool]]) -> dict[tuple[str, bool], torch.Tensor]:
        return {(k, v[5][0].item()) : v[0] for k, v in data_dict.items()}
    
    @torch.no_grad()
    def run(self):
        self.modelComposite.eval()  
        self.modelComposite.module.setState(NetState.PREDICTION)
        desc = lambda : "Prediction : {}".format(description(self.modelComposite))
        self.dataloader_prediction.dataset.load("Prediction")
        with tqdm.tqdm(iterable = enumerate(self.dataloader_prediction), leave=True, desc = desc(), total=len(self.dataloader_prediction), ncols=0) as batch_iter:
            for it, data_dict in batch_iter:
                with torch.amp.autocast('cuda', enabled=self.autocast):
                    input = self.getInput(data_dict)
                    for name, output in self.modelComposite(input, list(self.outsDataset.keys())):
                        self._predict_log(data_dict)
                        outDataset = self.outsDataset[name]
                        for i, (index, patch_augmentation, patch_index) in enumerate([(int(index), int(patch_augmentation), int(patch_index)) for index, patch_augmentation, patch_index in zip(list(data_dict.values())[0][1], list(data_dict.values())[0][2], list(data_dict.values())[0][3])]):
                            outDataset.addLayer(index, patch_augmentation, patch_index, output[i].cpu(), self.dataset)
                            if outDataset.isDone(index):
                                outDataset.write(index, self.dataset.getDatasetFromIndex(list(data_dict.keys())[0], index).name.split("/")[-1], outDataset.getOutput(index, self.dataset))

                    batch_iter.set_description(desc())
                    self.it += 1
        
    def _predict_log(self, data_dict : dict[str, tuple[torch.Tensor, int, int, int]]):
        measures = DistributedObject.getMeasure(self.world_size, self.global_rank, self.local_rank, {"" : self.modelComposite.module}, 1)
        
        if self.global_rank == 0:
            images_log = []
            if len(self.data_log):    
                for name, data_type in self.data_log.items():
                    if name in data_dict:
                        data_type[0](self.tb, "Prediction/{}".format(name), data_dict[name][0][:self.data_log[name][1]].detach().cpu().numpy(), self.it)
                    else:
                        images_log.append(name.replace(":", "."))

            for name, network in self.modelComposite.module.getNetworks().items():
                if network.measure is not None:
                    self.tb.add_scalars("Prediction/{}/Loss".format(name), {k : v[1] for k, v in measures["{}{}".format(name, "")][0].items()}, self.it)
                    self.tb.add_scalars("Prediction/{}/Metric".format(name), {k : v[1] for k, v in measures["{}{}".format(name, "")][1].items()}, self.it)
                if len(images_log):
                    for name, layer, _ in self.model.module.get_layers([v.to(0) for k, v in self.getInput(data_dict).items() if k[1]], images_log):
                        self.data_log[name][0](self.tb, "Prediction/{}".format(name), layer[:self.data_log[name][1]].detach().cpu().numpy(), self.it)

class ModelComposite(Network):

    def __init__(self, model: Network, nb_models: int, combine: Reduction):
        super().__init__(model.in_channels, model.optimizer, model.schedulers, model.outputsCriterionsLoader, model.patch, model.nb_batch_per_step, model.init_type, model.init_gain, model.dim)
        self.combine = combine
        for i in range(nb_models):
            self.add_module("Model_{}".format(i), copy.deepcopy(model), in_branch=[0], out_branch=["output_{}".format(i)])

    def load(self, state_dicts : list[dict[str, dict[str, torch.Tensor]]]):
        for i, state_dict in enumerate(state_dicts):
            self["Model_{}".format(i)].load(state_dict, init=False)
            self["Model_{}".format(i)].setName("{}_{}".format(self["Model_{}".format(i)].getName(), i))
            
    def forward(self, data_dict: dict[tuple[str, bool], torch.Tensor], output_layers: list[str] = []) -> list[tuple[str, torch.Tensor]]:
        result = {}
        for name, module in self.items():
            result[name] = module(data_dict, output_layers)
        
        aggregated = defaultdict(list)
        for module_outputs in result.values():
            for key, tensor in module_outputs:
                aggregated[key].append(tensor)

        final_outputs = []
        for key, tensors in aggregated.items():
            final_outputs.append((key, self.combine(torch.stack(tensors, dim=0))))

        return final_outputs
    

class Predictor(DistributedObject):

    @config("Predictor")
    def __init__(self, 
                    model: ModelLoader = ModelLoader(),
                    dataset: DataPrediction = DataPrediction(),
                    combine: str = "mean",
                    train_name: str = "name",
                    manual_seed : Union[int, None] = None,
                    gpu_checkpoints: Union[list[str], None] = None,
                    autocast : bool = False,
                    outsDataset: Union[dict[str, OutDatasetLoader], None] = {"default:Default" : OutDatasetLoader()},
                    images_log: list[str] = []) -> None:
        if os.environ["KONFAI_CONFIG_MODE"] != "Done":
            exit(0)
        super().__init__(train_name)
        self.manual_seed = manual_seed
        self.dataset = dataset
        self.combine_classpath = combine
        self.autocast = autocast

        self.model = model.getModel(train=False)
        self.it = 0
        self.outsDatasetLoader = outsDataset if outsDataset else {}
        self.outsDataset = {name.replace(":", ".") : value.getOutDataset(name) for name, value in self.outsDatasetLoader.items()}

        self.datasets_filename = []
        self.predict_path = PREDICTIONS_DIRECTORY()+self.name+"/"
        self.images_log = images_log
        for outDataset in self.outsDataset.values():
            self.datasets_filename.append(outDataset.filename)
            outDataset.filename = "{}{}".format(self.predict_path, outDataset.filename)
            
        
        self.gpu_checkpoints = gpu_checkpoints

    def _load(self) -> list[dict[str, dict[str, torch.Tensor]]]:
        model_paths = MODEL().split(":")
        state_dicts = []
        for model_path in model_paths:
            if model_path.startswith("https://"):
                try:
                    state_dicts.append(torch.hub.load_state_dict_from_url(url=model_path, map_location="cpu", check_hash=True))
                except:
                    raise Exception("Model : {} does not exist !".format(model_path)) 
            else:
                if model_path != "":
                    path = ""
                    name = model_path
                else:
                    if self.name.endswith(".pt"):
                        path = MODELS_DIRECTORY()+"/".join(self.name.split("/")[:-1])+"/StateDict/"
                        name = self.name.split("/")[-1]
                    else:
                        path = MODELS_DIRECTORY()+self.name+"/StateDict/"
                        name = sorted(os.listdir(path))[-1]
                if os.path.exists(path+name):
                    state_dicts.append(torch.load(path+name, weights_only=False))
                else:
                    raise Exception("Model : {} does not exist !".format(path+name))
        return state_dicts
    
    def setup(self, world_size: int):
        for dataset_filename in self.datasets_filename:
            path = self.predict_path +dataset_filename
            if os.path.exists(path):
                if os.environ["KONFAI_OVERWRITE"] != "True":
                    accept = builtins.input("The prediction {} already exists ! Do you want to overwrite it (yes,no) : ".format(path))
                    if accept != "yes":
                        return
                    
            if not os.path.exists(path):
                os.makedirs(path)

        shutil.copyfile(CONFIG_FILE(), self.predict_path+"Prediction.yml")

        
        self.model.init(self.autocast, State.PREDICTION, self.dataset.getGroupsDest())
        self.model.init_outputsGroup()
        self.model._compute_channels_trace(self.model, self.model.in_channels, None, self.gpu_checkpoints)
        
        modules = []
        for i,_,_ in self.model.named_ModuleArgsDict():
            modules.append(i)
        for output_group in self.outsDataset.keys():
            if output_group not in modules:
                raise PredictorError("The output group '{}' defined in 'outputsCriterions' does not correspond to any module in the model.".format(output_group),
                    "Available modules: {}".format(modules),
                    "Please check that the name matches exactly a submodule or output of your model architecture."
                )
        
        module, name = _getModule(self.combine_classpath, "konfai.predictor")
        if module == "konfai.predictor":
            combine = getattr(importlib.import_module(module), name)
        else:
            combine = config("{}.{}".format(KONFAI_ROOT(), self.combine_classpath))(getattr(importlib.import_module(module), name))(config = None)
       
        
        self.modelComposite = ModelComposite(self.model, len(MODEL().split(":")), combine)
        self.modelComposite.load(self._load())

        if len(list(self.outsDataset.keys())) == 0 and len([network for network in self.modelComposite.getNetworks().values() if network.measure is not None]) == 0:
            exit(0)
        
        self.size = (len(self.gpu_checkpoints)+1 if self.gpu_checkpoints else 1)
        self.dataloader = self.dataset.getData(world_size//self.size)
        for name, outDataset in self.outsDataset.items():
            outDataset.load(name.replace(".", ":"), list(self.dataset.datasets.values()), {src : dest for src, inner in self.dataset.groups_src.items() for dest in inner})
           
           
    def run_process(self, world_size: int, global_rank: int, local_rank: int, dataloaders: list[DataLoader]):
        modelComposite = Network.to(self.modelComposite, local_rank*self.size)
        modelComposite = DDP(modelComposite, static_graph=True) if torch.cuda.is_available() else CPU_Model(modelComposite)
        with _Predictor(world_size, global_rank, local_rank, self.autocast, self.predict_path, self.images_log, self.outsDataset, modelComposite, *dataloaders) as p:
            p.run()

        


import os
import ruamel.yaml
import inspect 
import collections
from copy import deepcopy
from typing import Union, Literal, get_origin, get_args, Any
import torch
from konfai import CONFIG_FILE
from konfai.utils.utils import ConfigError

yaml = ruamel.yaml.YAML()


class Config():

    def __init__(self, filename, key) -> None:
        self.filename = filename
        self.keys = key.split(".")

    def __enter__(self):
        if not os.path.exists(self.filename):
            result = input("Create a new config file ? [no,yes,interactive] : ")
            if result in ["yes", "interactive"]:
                os.environ["KONFAI_CONFIG_MODE"] = "interactive" if result == "interactive" else "default"
            else:
                exit(0)
            with open(self.filename, "w") as f:
                pass
                    
        self.yml = open(self.filename, 'r')
        self.data = yaml.load(self.yml)
        if self.data == None:
            self.data = {}
        
        self.config = self.data

        for key in self.keys:
            if self.config == None or key not in self.config:
                self.config = {key : {}}
            
            self.config = self.config[key]
        return self

    def createDictionary(self, data, keys, i) -> dict:
        if keys[i] not in data:
            data = {keys[i]: data}
        if i == 0:
            return data
        else:
            i -= 1
            return self.createDictionary(data, keys, i)
    
    def merge(self, dict1, dict2) -> dict:
        result = deepcopy(dict1)

        for key, value in dict2.items():
            if isinstance(value, collections.abc.Mapping):
                result[key] = self.merge(result.get(key, {}), value)
            else:
                if not dict2[key] == None:
                    result[key] = deepcopy(dict2[key])
        return result


    def __exit__(self, type, value, traceback) -> None:
        self.yml.close()
        if os.environ["KONFAI_CONFIG_MODE"] == "remove":
            if os.path.exists(CONFIG_FILE()):
                os.remove(CONFIG_FILE())
            return
        with open(self.filename, 'r') as yml:
            data = yaml.load(yml)
            if data == None:
                data = {}
        with open(self.filename, 'w') as yml:
            yaml.dump(self.merge(data, self.createDictionary(self.config, self.keys, len(self.keys)-1)), yml)
  
    @staticmethod
    def _getInput(name : str, default : str) ->  str:
        try:
            return input("{} [{}]: ".format(name, ",".join(default.split(":")[1:]) if len(default.split(":")) else ""))
        except:
            result = input("\nKeep a default configuration file ? (yes,no) : ")
            if result == "yes":
                os.environ["KONFAI_CONFIG_MODE"] = "default"
            else:
                os.environ["KONFAI_CONFIG_MODE"] = "remove"
                exit(0)
        return default.split(":")[1] if len(default.split(":")) > 1 else default

    @staticmethod
    def _getInputDefault(name : str, default : Union[str, None], isList : bool = False) -> Union[list[Union[str, None]], str, None]:
        if isinstance(default, str) and (default == "default" or (len(default.split(":")) > 1 and default.split(":")[0] == "default")):
            if os.environ["KONFAI_CONFIG_MODE"] == "interactive":
                if isList:
                    list_tmp = []
                    key_tmp = "OK"
                    while (key_tmp != "!" and key_tmp != " ") and os.environ["KONFAI_CONFIG_MODE"] == "interactive":
                        key_tmp = Config._getInput(name, default)
                        if (key_tmp != "!" and key_tmp != " "):
                            if key_tmp == "":
                                key_tmp = default.split(":")[1] if len(default.split(":")) > 1 else default
                            list_tmp.append(key_tmp)
                    return list_tmp
                else:
                    value = Config._getInput(name, default)
                    if value == "":
                        return default.split(":")[1] if len(default.split(":")) > 1 else default 
                    else: 
                        return value
            else:
                default = default.split(":")[1] if len(default.split(":")) > 1 else default
        return  [default] if isList else default

    def getValue(self, name, default) -> object:  
        if name in self.config and self.config[name] is not None:
            value = self.config[name]
            if value == None:
                value = default
            value_config = value
        else:
            value = Config._getInputDefault(name, default if default != inspect._empty else None)
            
            value_config = value
            if type(value_config) == tuple:
                value_config = list(value)
                
            if type(value_config) == list:
                list_tmp = [] 
                for key in value_config:
                    list_tmp.extend(Config._getInputDefault(name, key, isList=True))
                

                value = list_tmp
                value_config = list_tmp

            if type(value) == dict:
                key_tmp = []
                
                value_config = {}
                dict_value = {}
                for key in value:
                    key_tmp.extend(Config._getInputDefault(name, key, isList=True))
                for key in key_tmp:
                    if key in value:
                        value_tmp = value[key]
                    else:
                        value_tmp = next(v for k,v in value.items() if "default" in k)

                    value_config[key] = None
                    dict_value[key] = value_tmp
                value = dict_value
        if isinstance(self.config, str):
            os.environ['KONFAI_CONFIG_VARIABLE'] = "True"
            return None

        self.config[name] = value_config if value_config is not None else "None"
        if value == "None":
            value = None
        return value
                    
def config(key : Union[str, None] = None):
    def decorator(function):
        def new_function(*args, **kwargs):
            if "config" in kwargs:
                filename =  kwargs["config"]
                if filename == None:
                    filename = os.environ['KONFAI_CONFIG_FILE']
                else:
                    os.environ['KONFAI_CONFIG_FILE'] = filename
                key_tmp =  kwargs["DL_args"]+("."+key if key is not None else "") if "DL_args" in kwargs else key
                without =  kwargs["DL_without"] if "DL_without" in kwargs else []
                os.environ['KONFAI_CONFIG_PATH'] = key_tmp
                with Config(filename, key_tmp) as config:
                    os.environ['KONFAI_CONFIG_VARIABLE'] = "False"
                    kwargs = {} 
                    for param in list(inspect.signature(function).parameters.values())[len(args):]:
                        if param.name in without:
                            continue
                        
                        annotation = param.annotation

                        # --- support Literal ---
                        if get_origin(annotation) is Literal:
                            allowed_values = get_args(annotation)
                            default_value = param.default if param.default != inspect._empty else allowed_values[0]
                            value = config.getValue(param.name, f"default:{default_value}")
                            if value not in allowed_values:
                                raise ConfigError(
                                    f"Invalid value '{value}' for parameter '{param.name} expected one of: {allowed_values}."
                                )
                            kwargs[param.name] = value
                            continue
                        if str(annotation).startswith("typing.Union") or str(annotation).startswith("typing.Optional"):
                            for i in annotation.__args__:
                                annotation = i
                                break
                        
                        if not annotation == inspect._empty:
                            if annotation not in [int, str, bool, float, torch.Tensor]:
                                if str(annotation).startswith("list") or str(annotation).startswith("tuple") or str(annotation).startswith("typing.Tuple") or str(annotation).startswith("typing.List") or str(annotation).startswith("typing.Sequence"):
                                    elem_type = annotation.__args__[0]
                                    values = config.getValue(param.name, param.default)
                                    if getattr(elem_type, '__origin__', None) is Union:
                                        valid_types = elem_type.__args__
                                        result = []
                                        for v in values:
                                            for t in valid_types:
                                                try:
                                                    if t == torch.Tensor and not isinstance(v, torch.Tensor):
                                                        v = torch.tensor(v)
                                                    result.append(t(v) if t != torch.Tensor else v)
                                                    break
                                                except Exception:
                                                    continue
                                        kwargs[param.name] = result

                                    elif annotation.__args__[0] in [int, str, bool, float]:
                                        values = config.getValue(param.name, param.default)
                                        kwargs[param.name] = values
                                    else:
                                        raise ConfigError("Config: The config only supports types : config(Object), int, str, bool, float, list[int], list[str], list[bool], list[float], dict[str, Object]")
                                elif str(annotation).startswith("dict"):
                                    if annotation.__args__[0] == str:
                                        values = config.getValue(param.name, param.default)
                                        if values is not None and annotation.__args__[1] not in [int, str, bool, float, Any]:
                                            try:
                                                kwargs[param.name] = {value : annotation.__args__[1](config = filename, DL_args = key_tmp+"."+param.name+"."+value) for value in values}
                                            except ValueError as e:
                                                raise ValueError(e)
                                            except Exception as e:
                                                raise ConfigError("{} {}".format(values, e))
                                        else:

                                            kwargs[param.name] = values
                                    else: 
                                        raise ConfigError("Config: The config only supports types : config(Object), int, str, bool, float, list[int], list[str], list[bool], list[float], dict[str, Object]")
                                else:
                                    try:
                                        kwargs[param.name] = annotation(config = filename, DL_args = key_tmp)
                                    except Exception as e:
                                        raise ConfigError("Failed to instantiate {} with type {}, error {} ".format(param.name, annotation, e))
                                    
                                    if os.environ['KONFAI_CONFIG_VARIABLE'] == "True":
                                        os.environ['KONFAI_CONFIG_VARIABLE'] = "False"
                                        kwargs[param.name] = None
                            else:
                                kwargs[param.name] = config.getValue(param.name, param.default)
                        elif param.name != "self":
                            kwargs[param.name] = config.getValue(param.name, param.default)
            result = function(*args, **kwargs)
            return result
        return new_function
    return decorator



import SimpleITK as sitk
import h5py
from abc import ABC, abstractmethod
import numpy as np
from typing import Any, Union
import copy
import torch
import os

from lxml import etree
import csv
from konfai import DATE

class Plot():

    def __init__(self, root: etree.ElementTree) -> None:
        self.root = root

    def _explore(root, result, label):
        if len(root) == 0:
            for attribute in root.attrib:
                result["attrib:"+label+":"+attribute] = root.attrib[attribute]
            if root.text is not None:
                result[label] = np.fromstring(root.text, sep = ",").astype('double')
        else:
            for node in root:
                Plot._explore(node, result, label+":"+node.tag)

    def getNodes(root, path = None, id = None):
        nodes = []
        if path != None:
            path = path.split(":")
            for node_name in path:
                node = root.find(node_name)
                if node != None:
                    root = node
                else:
                    break
        if id != None:
            for node in root.findall(".//"+id):
                nodes.append(node)
        else:
            nodes.append(root)                
        return nodes

    def read(root, path = None, id = None):
        result = dict()
        for node in Plot.getNodes(root, path, id):
            Plot._explore(node, result, etree.ElementTree(root).getpath(node))    
        return result

    def _extract(self, ids = [], patients = []):
        result = dict()
        if len(patients) == 0:
            if len(ids) == 0:
                result.update(Plot.read(self.root,None, None))
            else:
                for id in ids:
                    result.update(Plot.read(self.root, None, id))
        else:
            for path in patients:
                if len(ids) == 0:
                    result.update(Plot.read(self.root, path, None))
                else:
                    for id in ids:
                        result.update(Plot.read(self.root, path, id))
        return result

    def getErrors(self, ids = [], patients = []):
        results = self._extract(ids=ids, patients=patients)
        errors = {k: v for k, v in results.items() if not k.startswith("attrib:")}
        results : dict[str, dict[str, np.ndarray]]= {}
        for key, error in errors.items():
            patient = key.replace("/",":").split(":")[2]
            k = key.replace("/",":").split(":")[-1]
            err = np.linalg.norm(error.reshape(int(error.shape[0]/3),3), ord=2, axis=1)
            if patient not in results:
                results[patient] = {k : err}
            else:
                results[patient].update({k : err})
        return results
    
    def statistic_attrib(self, ids = [], patients = [], type: str = "HD95Mean"):
        results = self._extract(ids=ids, patients=patients)
        
        errors = {k.replace("attrib:", ""): float(v) for k, v in results.items() if type in k}
        
        values = {key : np.array([]) for key in ids}
        for key, error in errors.items():
            k = key.replace("/",":").split(":")[-2]
            values[k] = np.append(values[k], error)
        
        for k in values:
            values[k] = np.mean(values[k])
        print(values)
        return values
    
    def statistic_parameter(self, ids = [], patients = []):
        results = self._extract(ids=ids, patients=patients)
        errors = {k.replace("attrib:", "").replace(":Time", "") : np.load("./Results/{}/{}.npy".format(k.split("/")[3].split(":")[0], k.split("/")[2])) for k in results.keys()}
        
        norms = {key : np.array([]) for key in ids}
        max = 0
        for key, error in errors.items():
            if max < int(error.shape[0]/3): 
                max = int(error.shape[0]/3)
        
        for key, error in errors.items():
            k = key.replace("/",":").split(":")[-1]
            norms[k] = np.append(norms[k], np.linalg.norm(error.reshape(int(error.shape[0]/3),3), ord=2, axis=1))
            v = np.linalg.norm(error.reshape(int(error.shape[0]/3),3), ord=2, axis=1)

            print(key, "{} {} {} {} {}".format(np.round(np.mean(v), 2), np.round(np.std(v), 2), np.round(np.quantile(v, 0.25), 2), np.round(np.quantile(v, 0.5), 2), np.round(np.quantile(v, 0.75), 2))) 
        results = {}
        for key, values in norms.items():
            if key == "Rigid":
                results.update({key : values})
            else:    
                try:
                    name = "{}".format("_".join(key.split("_"))[:-1])
                    it = int(key.split("_")[-1])
                except:
                    name = "{}".format(key.split("-")[0])
                    it = int(key.split("-")[-1])
                
                if name in results:
                    results[name].update({it : values})
                else:
                    results.update({name : {it : values}})
        
        r = []
        for key, values in norms.items():
            #r.append("{} $\pm$ {}".format(np.round(np.mean(values), 2), np.round(np.std(values), 2))) 
            r.append("{} {} {}".format(np.round(np.quantile(values, 0.25), 2), np.round(np.quantile(values, 0.5),2), np.round(np.quantile(values, 0.75), 2)))
            #r.append("{} $\pm$ {}".format(np.round(np.quantile(values, 0.5), 2), np.round(np.quantile(values, 0.75)-np.quantile(values, 0.25), 2)))
        print(" & ".join(r))

    def statistic(self, ids = [], patients = []):
        results = self._extract(ids=ids, patients=patients)
        #errors = {k.replace("attrib:", "").replace(":Time", "") : np.load("./Dataset/{}/{}.npy".format(k.split("/")[3].split(":")[0], k.split("/")[2])) for k in results.keys()}
        errors = {k: v for k, v in results.items() if not k.startswith("attrib:")}
        print(errors)
        norms = {key : np.array([]) for key in ids}
        max = 0
        for key, error in errors.items():
            if max < int(error.shape[0]/3): 
                max = int(error.shape[0]/3)
        
        for key, error in errors.items():
            k = key.replace("/",":").split(":")[-1]
            norms[k] = np.append(norms[k], np.linalg.norm(error.reshape(int(error.shape[0]/3),3), ord=2, axis=1))
            v = np.linalg.norm(error.reshape(int(error.shape[0]/3),3), ord=2, axis=1)
            print(key, (np.mean(v), np.std(v), np.quantile(v, 0.25), np.quantile(v, 0.5), np.quantile(v, 0.75)) )
        results = {}
        """for key, values in norms.items():
            if key == "Rigid":
                results.update({key : values})
            else:    
                try:
                    name = "{}".format("_".join(key.split("_"))[:-1])
                    it = int(key.split("_")[-1])
                except:
                    name = "{}".format(key.split("-")[0])
                    it = int(key.split("-")[-1])
                
                if name in results:
                    results[name].update({it : values})
                else:
                    results.update({name : {it : values}})"""


        print({key: (np.mean(values), np.std(values), np.quantile(values, 0.25), np.quantile(values, 0.5), np.quantile(values, 0.75)) for key, values in norms.items()})
        return results
    
    def plot(self, ids = [], patients = [], labels = [], colors = None):

        import matplotlib.pyplot as pyplot
        results = self._extract(ids=ids, patients=patients)

        attrs = {k: v for k, v in results.items() if k.startswith("attrib:")}
        errors = {k: v for k, v in results.items() if not k.startswith("attrib:")}

        patients = set()
        max = 0
        for key, error in errors.items():
            patients.add(key.replace("/",":").split(":")[2])
            if max < int(error.shape[0]/3): 
                max = int(error.shape[0]/3)
        patients = sorted(patients)

        norms = {patient : np.array([]) for patient in patients}
        markups = {patient : np.array([]) for patient in patients}
        series = list()
        for key, error in errors.items():
            patient = key.replace("/",":").split(":")[2]
            markup = np.full((max,3), np.nan)
            markup[0:int(error.shape[0]/3), :] = error.reshape(int(error.shape[0]/3),3)

            markups[patient] = np.append(markups[patient], markup)
            norms[patient] = np.append(norms[patient], np.linalg.norm(markup, ord=2, axis=1))

        if len(labels) == 0:
            labels = list(set([k.split("/")[-1] for k in errors.keys()]))

        for label in labels:
            series = series+[label]*max
        import pandas as pd
        df = pd.DataFrame(dict([(k,pd.Series(v)) for k, v in norms.items()]))
        df['Categories'] = pd.Series(series)
        
        bp = df.boxplot(by='Categories', color="black", figsize=(12,8), notch=True,layout=(1,len(patients)), fontsize=18, rot=0, patch_artist = True, return_type='both',  widths=[0.5]*len(labels))

        color_pallet = {"b" : "paleturquoise", "g" : "lightgreen"}
        if colors == None:
            colors = ["b"] * len(patients)
        pyplot.suptitle('')
        it_1 = 0
        for index, (ax,row)  in bp.items():
            ax.set_xlabel('')
            ax.set_ylim(ymin=0)
            ax.set_ylabel("TRE (mm)", fontsize=18)
            ax.set_yticks([0,1,2,3,4,5,6,7,8,9,10,15,20,25])  # Set label locations.
            for i,object in enumerate(row["boxes"]):
                object.set_edgecolor("black")
                object.set_facecolor(color_pallet[colors[i]])
                object.set_alpha(0.7)
                object.set_linewidth(1.0)
            
            for i,object in enumerate(row["medians"]):
                object.set_color("indianred")
                xy = object.get_xydata()
                object.set_linewidth(2.0)
                it_1+=1
        return self
    
    def show(self):
        import matplotlib.pyplot as pyplot
        pyplot.show()

class Attribute(dict[str, Any]):

    def __init__(self, attributes : dict[str, Any] = {}) -> None:
        super().__init__()
        for k, v in attributes.items():
            super().__setitem__(copy.deepcopy(k), copy.deepcopy(v))
    
    def __getitem__(self, key: str) -> Any:
        i = len([k for k in super().keys() if k.startswith(key)])
        if i > 0 and "{}_{}".format(key, i-1) in super().keys():   
            return str(super().__getitem__("{}_{}".format(key, i-1)))
        else:
            raise NameError("{} not in cache_attribute".format(key))

    def __setitem__(self, key: str, value: Any) -> None:
        if "_" not in key:
            i = len([k for k in super().keys() if k.startswith(key)])
            result = None
            if isinstance(value, torch.Tensor):
                result = str(value.numpy())
            else:
                result = str(value)
            result = result.replace('\n', '')
            super().__setitem__("{}_{}".format(key, i), result)
        else:
            result = None
            if isinstance(value, torch.Tensor):
                result = str(value.numpy())
            else:
                result = str(value)
            result = result.replace('\n', '')
            super().__setitem__(key, result)

    def pop(self, key: str) -> Any:
        i = len([k for k in super().keys() if k.startswith(key)])
        if i > 0 and "{}_{}".format(key, i-1) in super().keys():   
            return super().pop("{}_{}".format(key, i-1))
        else:
            raise NameError("{} not in cache_attribute".format(key))

    def get_np_array(self, key) -> np.ndarray:
        return np.fromstring(self[key][1:-1], sep=" ", dtype=np.double)
    
    def get_tensor(self, key) -> torch.Tensor:
        return torch.tensor(self.get_np_array(key)).to(torch.float32)
    
    def pop_np_array(self, key):
        return np.fromstring(self.pop(key)[1:-1], sep=" ", dtype=np.double)
    
    def pop_tensor(self, key) -> torch.Tensor:
        return torch.tensor(self.pop_np_array(key))
    
    def __contains__(self, key: str) -> bool:
        return len([k for k in super().keys() if k.startswith(key)]) > 0
    
    def isInfo(self, key: str, value: str) -> bool:
        return key in self and self[key] == value

def isAnImage(attributes: Attribute):
    return "Origin" in attributes and "Spacing" in attributes and "Direction" in attributes

def data_to_image(data : np.ndarray, attributes: Attribute) -> sitk.Image:
    if not isAnImage(attributes):
        raise NameError("Data is not an image")
    if data.shape[0] == 1:
        image = sitk.GetImageFromArray(data[0])
    else:
        data = data.transpose(tuple([i+1 for i in range(len(data.shape)-1)]+[0]))
        image = sitk.GetImageFromArray(data, isVector=True)
    image.SetOrigin(attributes.get_np_array("Origin").tolist())
    image.SetSpacing(attributes.get_np_array("Spacing").tolist())
    image.SetDirection(attributes.get_np_array("Direction").tolist())
    return image

def image_to_data(image: sitk.Image) -> tuple[np.ndarray, Attribute]:
    attributes = Attribute()
    attributes["Origin"] = np.asarray(image.GetOrigin())
    attributes["Spacing"] = np.asarray(image.GetSpacing())
    attributes["Direction"] = np.asarray(image.GetDirection())
    data = sitk.GetArrayFromImage(image)

    if image.GetNumberOfComponentsPerPixel() == 1:
        data = np.expand_dims(data, 0)
    else:
        data = np.transpose(data, (len(data.shape)-1, *[i for i in range(len(data.shape)-1)]))
    return data, attributes

class Dataset():

    class AbstractFile(ABC):

        def __init__(self) -> None:
            pass
        
        def __enter__(self):
            pass

        def __exit__(self, type, value, traceback):
            pass

        @abstractmethod
        def file_to_data(self):
            pass

        @abstractmethod
        def data_to_file(self):
            pass

        @abstractmethod
        def getNames(self, group: str) -> list[str]:
            pass
        
        @abstractmethod
        def getGroup(self) -> list[str]:
            pass

        @abstractmethod
        def isExist(self, group: str, name: Union[str, None] = None) -> bool:
            pass
        
        @abstractmethod
        def getInfos(self, group: Union[str, None], name: str) -> tuple[list[int], Attribute]:
            pass

    class H5File(AbstractFile):

        def __init__(self, filename: str, read: bool) -> None:
            self.h5: Union[h5py.File, None] = None
            self.filename = filename
            if not self.filename.endswith(".h5"):
                self.filename += ".h5"
            self.read = read

        def __enter__(self):
            args = {}
            if self.read:
                self.h5 = h5py.File(self.filename, 'r', **args)
            else:
                if not os.path.exists(self.filename):
                    if len(self.filename.split("/")) > 1 and not os.path.exists("/".join(self.filename.split("/")[:-1])):
                        os.makedirs("/".join(self.filename.split("/")[:-1]))
                    self.h5 = h5py.File(self.filename, 'w', **args)
                else: 
                    self.h5 = h5py.File(self.filename, 'r+', **args)
                self.h5.attrs["Date"] = DATE()
            self.h5.__enter__()
            return self.h5
    
        def __exit__(self, type, value, traceback):
            if self.h5 is not None:
                self.h5.close()
        
        def file_to_data(self, groups: str, name: str) -> tuple[np.ndarray, Attribute]:
            dataset = self._getDataset(groups, name)
            data = np.zeros(dataset.shape, dataset.dtype)
            dataset.read_direct(data)
            return data, Attribute({k : str(v) for k, v in dataset.attrs.items()})

        def data_to_file(self, name : str, data : Union[sitk.Image, sitk.Transform, np.ndarray], attributes : Union[Attribute, None] = None) -> None:
            if attributes is None:
                attributes = Attribute()
            if isinstance(data, sitk.Image):
                data, attributes_tmp = image_to_data(data)
                attributes.update(attributes_tmp)
            elif isinstance(data, sitk.Transform):
                transforms = []
                if isinstance(data, sitk.CompositeTransform):
                    for i in range(data.GetNumberOfTransforms()):
                        transforms.append(data.GetNthTransform(i))    
                else:
                    transforms.append(data)
                datas = []
                for i, transform in enumerate(transforms):
                    if isinstance(transform, sitk.Euler3DTransform):
                        transform_type = "Euler3DTransform_double_3_3"
                    if isinstance(transform, sitk.AffineTransform):
                        transform_type = "AffineTransform_double_3_3"
                    if isinstance(transform, sitk.BSplineTransform):
                        transform_type = "BSplineTransform_double_3_3"
                    attributes["{}:Transform".format(i)] = transform_type
                    attributes["{}:FixedParameters".format(i)] = transform.GetFixedParameters()

                    datas.append(np.asarray(transform.GetParameters()))
                data = np.asarray(datas)

            h5_group = self.h5
            if len(name.split("/")) > 1:
                group = "/".join(name.split("/")[:-1])
                if group not in self.h5:
                    self.h5.create_group(group)
                h5_group = self.h5[group]

            name = name.split("/")[-1]
            if name in h5_group:
                del h5_group[name]

            dataset = h5_group.create_dataset(name, data=data, dtype=data.dtype, chunks=None)
            dataset.attrs.update({k : str(v) for k, v in attributes.items()})
        
        def isExist(self, group: str, name: Union[str, None] = None) -> bool:
            if group in self.h5:
                if isinstance(self.h5[group], h5py.Dataset):
                    return True
                elif name is not None:
                    return name in self.h5[group]
                else:
                    return False
            return False

        def getNames(self, groups: str, h5_group: h5py.Group = None) -> list[str]:
            names = []
            if h5_group is None:
                h5_group = self.h5
            group = groups.split("/")[0]
            if group == "":
                names = [dataset.name.split("/")[-1] for dataset in h5_group.values() if isinstance(dataset, h5py.Dataset)]
            elif group == "*":
                for k in h5_group.keys():
                    if isinstance(h5_group[k], h5py.Group):
                        names.extend(self.getNames("/".join(groups.split("/")[1:]), h5_group[k]))
            else:
                if group in h5_group:
                    names.extend(self.getNames("/".join(groups.split("/")[1:]), h5_group[group]))
            return names
        
        def getGroup(self):
            return self.h5.keys()
        
        def _getDataset(self, groups: str, name: str, h5_group: h5py.Group = None) -> h5py.Dataset:
            if h5_group is None:
                h5_group = self.h5
            if groups != "":
                group = groups.split("/")[0]
            else:
                group = ""
            result = None
            if group == "":
                if name in h5_group:
                    result = h5_group[name]
            elif group == "*":
                for k in h5_group.keys():
                    if isinstance(h5_group[k], h5py.Group):
                        result_tmp = self._getDataset("/".join(groups.split("/")[1:]), name, h5_group[k])
                        if result_tmp is not None:
                            result = result_tmp
            else:
                if group in h5_group:
                    result_tmp = self._getDataset("/".join(groups.split("/")[1:]), name, h5_group[group])
                    if result_tmp is not None:
                        result = result_tmp
            return result
        
        def getInfos(self, groups: str, name: str) -> tuple[list[int], Attribute]:
            dataset = self._getDataset(groups, name)
            return (dataset.shape, Attribute({k : str(v) for k, v in dataset.attrs.items()}))
        
    class SitkFile(AbstractFile):

        def __init__(self, filename: str, read: bool, format: str) -> None:
            self.filename = filename
            self.read = read
            self.format = format
     
        def file_to_data(self, group: str, name: str) -> tuple[np.ndarray, Attribute]:
            attributes = Attribute()
            if os.path.exists("{}{}.{}".format(self.filename, name, self.format)):
                image = sitk.ReadImage("{}{}.{}".format(self.filename, name, self.format))
                data, attributes_tmp = image_to_data(image)
                attributes.update(attributes_tmp)
            elif os.path.exists("{}{}.itk.txt".format(self.filename, name)): 
                data = sitk.ReadTransform("{}{}.itk.txt".format(self.filename, name))
                transforms = []
                if isinstance(data, sitk.CompositeTransform):
                    for i in range(data.GetNumberOfTransforms()):
                        transforms.append(data.GetNthTransform(i))    
                else:
                    transforms.append(data)
                datas = []
                for i, transform in enumerate(transforms):
                    if isinstance(transform, sitk.Euler3DTransform):
                        transform_type = "Euler3DTransform_double_3_3"
                    if isinstance(transform, sitk.AffineTransform):
                        transform_type = "AffineTransform_double_3_3"
                    if isinstance(transform, sitk.BSplineTransform):
                        transform_type = "BSplineTransform_double_3_3"
                    attributes["{}:Transform".format(i)] = transform_type
                    attributes["{}:FixedParameters".format(i)] = transform.GetFixedParameters()

                    datas.append(np.asarray(transform.GetParameters()))
                data = np.asarray(datas)
            elif os.path.exists("{}{}.fcsv".format(self.filename, name)):
                with open("{}{}.fcsv".format(self.filename, name), newline="") as csvfile:
                    reader = csv.reader(filter(lambda row: row[0]!='#', csvfile))
                    lines = list(reader)
                    data = np.zeros((len(list(lines)), 3), dtype=np.double)
                    for i, row in enumerate(lines):
                        data[i] = np.array(row[1:4], dtype=np.double)
                    csvfile.close()
            elif os.path.exists("{}{}.xml".format(self.filename, name)):
                with open("{}{}.xml".format(self.filename, name), 'rb') as xml_file:
                    result = etree.parse(xml_file, etree.XMLParser(remove_blank_text=True)).getroot()
                    xml_file.close()
                    return result
            elif os.path.exists("{}{}.vtk".format(self.filename, name)):
                import vtk
                vtkReader = vtk.vtkPolyDataReader()
                vtkReader.SetFileName("{}{}.vtk".format(self.filename, name))
                vtkReader.Update()
                data = []
                points = vtkReader.GetOutput().GetPoints()
                num_points = points.GetNumberOfPoints()
                for i in range(num_points):
                    data.append(list(points.GetPoint(i)))
                data = np.asarray(data)
            elif os.path.exists("{}{}.npy".format(self.filename, name)):
                data = np.load("{}{}.npy".format(self.filename, name))
            return data, attributes
        
        def is_vtk_polydata(self, obj):
            try:
                import vtk
                return isinstance(obj, vtk.vtkPolyData)
            except ImportError:
                return False
                
        def data_to_file(self, name : str, data : Union[sitk.Image, sitk.Transform, np.ndarray], attributes : Attribute = Attribute()) -> None:
            if not os.path.exists(self.filename):
                os.makedirs(self.filename)
            if isinstance(data, sitk.Image):
                for k, v in attributes.items():
                    data.SetMetaData(k, v)
                sitk.WriteImage(data, "{}{}.{}".format(self.filename, name, self.format))
            elif isinstance(data, sitk.Transform):
                sitk.WriteTransform(data, "{}{}.itk.txt".format(self.filename, name))
            elif self.is_vtk_polydata(data):
                import vtk
                vtkWriter = vtk.vtkPolyDataWriter()
                vtkWriter.SetFileName("{}{}.vtk".format(self.filename, name))
                vtkWriter.SetInputData(data)
                vtkWriter.Write()
            elif isAnImage(attributes):   
                self.data_to_file(name, data_to_image(data, attributes), attributes)
            elif (len(data.shape) == 2 and data.shape[1] == 3 and data.shape[0] > 0):
                data = np.round(data, 4)
                with open("{}{}.fcsv".format(self.filename, name), 'w') as f:
                    f.write("# Markups fiducial file version = 4.6\n# CoordinateSystem = 0\n# columns = id,x,y,z,ow,ox,oy,oz,vis,sel,lock,label,desc,associatedNodeID\n")
                    for i in range(data.shape[0]):
                        f.write("vtkMRMLMarkupsFiducialNode_"+str(i+1)+","+str(data[i, 0])+","+str(data[i, 1])+","+str(data[i, 2])+",0,0,0,1,1,1,0,F-"+str(i+1)+",,vtkMRMLScalarVolumeNode1\n")
                    f.close()
            elif "path" in attributes:
                if os.path.exists("{}{}.xml".format(self.filename, name)):
                    with open("{}{}.xml".format(self.filename, name), 'rb') as xml_file:
                        root = etree.parse(xml_file, etree.XMLParser(remove_blank_text=True)).getroot()
                        xml_file.close()
                else:
                    root = etree.Element(name)
                node = root
                path = attributes["path"].split(':')

                for node_name in path:
                    node_tmp = node.find(node_name)
                    if node_tmp == None:
                        node_tmp = etree.SubElement(node, node_name)
                        node.append(node_tmp)
                    node = node_tmp
                if attributes != None:
                    for attribute_tmp in attributes.keys():
                        attribute = "_".join(attribute_tmp.split("_")[:-1])
                        if attribute != "path":
                            node.set(attribute, attributes[attribute])
                if data.size > 0:
                    node.text = ", ".join(map(str, data.flatten())) #np.array2string(data, separator=',')[1:-1].replace('\n','')
                with open("{}{}.xml".format(self.filename, name), 'wb') as f:
                    f.write(etree.tostring(root, pretty_print=True, encoding='utf-8'))
                    f.close()
            else:
                np.save("{}{}.npy".format(self.filename, name), data)
                 
        def isExist(self, group: str, name: Union[str, None] = None) -> bool:
            return os.path.exists("{}{}.{}".format(self.filename, group, self.format)) or os.path.exists("{}{}.itk.txt".format(self.filename, group)) or os.path.exists("{}{}.fcsv".format(self.filename, group)) or os.path.exists("{}{}.npy".format(self.filename, group))
        
        def getNames(self, group: str) -> list[str]:
            raise NotImplementedError()
        
        def getGroup(self):
            raise NotImplementedError()
        
        def getInfos(self, group: str, name: str) -> tuple[list[int], Attribute]:
            attributes = Attribute()
            if os.path.exists("{}{}{}.{}".format(self.filename, group if group is not None else "", name, self.format)):
                file_reader = sitk.ImageFileReader()
                file_reader.SetFileName("{}{}{}.{}".format(self.filename, group if group is not None else "", name, self.format))
                file_reader.ReadImageInformation()
                attributes["Origin"] = np.asarray(file_reader.GetOrigin())
                attributes["Spacing"] = np.asarray(file_reader.GetSpacing())
                attributes["Direction"] = np.asarray(file_reader.GetDirection())
                for k in file_reader.GetMetaDataKeys():
                    attributes[k] = file_reader.GetMetaData(k)
                size = list(file_reader.GetSize())
                if len(size) == 3:
                    size = list(reversed(size))
                size = [file_reader.GetNumberOfComponents()]+size
            else:
                data, attributes = self.file_to_data(group if group is not None else "", name)
                size = data.shape
            return tuple(size), attributes

    class File(ABC):

        def __init__(self, filename: str, read: bool, format: str) -> None:
            self.filename = filename
            self.read = read
            self.file = None
            self.format = format

        def __enter__(self):
            if self.format == "h5":
                self.file = Dataset.H5File(self.filename, self.read)
            else:
                self.file = Dataset.SitkFile(self.filename+"/", self.read, self.format)
            self.file.__enter__()
            return self.file

        def __exit__(self, type, value, traceback):
            self.file.__exit__(type, value, traceback)

    def __init__(self, filename : str, format: str) -> None:
        if format != "h5" and not filename.endswith("/"):
            filename = "{}/".format(filename)
        self.is_directory = filename.endswith("/") 
        self.filename = filename
        self.format = format
        
    def write(self, group : str, name : str, data : Union[sitk.Image, sitk.Transform, np.ndarray], attributes : Attribute = Attribute()):
        if self.is_directory:
            if not os.path.exists(self.filename):
                os.makedirs(self.filename)
        if self.is_directory:
            s_group = group.split("/")
            if len(s_group) > 1:
                subDirectory = "/".join(s_group[:-1])
                name = "{}/{}".format(subDirectory, name)
                group = s_group[-1]
            with Dataset.File("{}{}".format(self.filename, name), False, self.format) as file:
                file.data_to_file(group, data, attributes)
        else:
            with Dataset.File(self.filename, False, self.format) as file:
                file.data_to_file("{}/{}".format(group, name), data, attributes)
    
    def readData(self, groups : str, name : str) -> tuple[np.ndarray, Attribute]:
        if not os.path.exists(self.filename):
            raise NameError("Dataset {} not found".format(self.filename))
        if self.is_directory:
            for subDirectory in self._getSubDirectories(groups):
                group = groups.split("/")[-1]
                if os.path.exists("{}{}{}{}".format(self.filename, subDirectory, name, ".h5" if self.format == "h5" else "")):
                    with Dataset.File("{}{}{}".format(self.filename, subDirectory, name), False, self.format) as file:
                        result = file.file_to_data("", group)
        else:
            with Dataset.File(self.filename, False, self.format) as file:
                result = file.file_to_data(groups, name)
        return result
    
    def readTransform(self, group : str, name : str) -> sitk.Transform:
        if not os.path.exists(self.filename):
            raise NameError("Dataset {} not found".format(self.filename))
        transformParameters, attribute = self.readData(group, name)
        transforms_type = [v for k, v in attribute.items() if k.endswith(":Transform_0")]
        transforms = []
        for i, transform_type in enumerate(transforms_type):
            if transform_type == "Euler3DTransform_double_3_3":
                transform = sitk.Euler3DTransform()
            if transform_type == "AffineTransform_double_3_3":
                transform = sitk.AffineTransform(3)
            if transform_type == "BSplineTransform_double_3_3":
                transform = sitk.BSplineTransform(3)
            transform.SetFixedParameters(eval(attribute["{}:FixedParameters".format(i)]))
            transform.SetParameters(tuple(transformParameters[i]))
            transforms.append(transform)
        return sitk.CompositeTransform(transforms) if len(transforms) > 1 else transforms[0]

    def readImage(self, group : str, name : str):
         data, attribute = self.readData(group, name)
         return data_to_image(data, attribute)
            
    def getSize(self, group: str) -> int:
        return len(self.getNames(group))
    
    def isGroupExist(self, group: str) -> bool:
        return self.getSize(group) > 0
    
    def isDatasetExist(self, group: str, name: str) -> bool:
        return name in self.getNames(group)
    
    def _getSubDirectories(self, groups: str, subDirectory: str = ""):
        group = groups.split("/")[0]
        subDirectories = []
        if len(groups.split("/")) == 1:
            subDirectories.append(subDirectory)
        elif group == "*":
            for k in os.listdir("{}{}".format(self.filename, subDirectory)):
                if not os.path.isfile("{}{}{}".format(self.filename, subDirectory, k)):
                    subDirectories.extend(self._getSubDirectories("/".join(groups.split("/")[1:]), "{}{}/".format(subDirectory , k)))
        else:
            subDirectory = "{}{}/".format(subDirectory, group)
            if os.path.exists("{}{}".format(self.filename, subDirectory)):
                subDirectories.extend(self._getSubDirectories("/".join(groups.split("/")[1:]), subDirectory))
        return subDirectories

    def getNames(self, groups: str, index: Union[list[int], None] = None) -> list[str]:
        names = []
        if self.is_directory:
            for subDirectory in self._getSubDirectories(groups):
                group = groups.split("/")[-1]
                if os.path.exists("{}{}".format(self.filename, subDirectory)):
                    for name in sorted(os.listdir("{}{}".format(self.filename, subDirectory))):
                        if os.path.isfile("{}{}{}".format(self.filename, subDirectory, name)) or self.format != "h5":
                            with Dataset.File("{}{}{}".format(self.filename, subDirectory, name), True, self.format) as file:
                                if file.isExist(group):
                                    names.append(name.replace(".h5", "") if self.format == "h5" else name)
        else:
            with Dataset.File(self.filename, True, self.format) as file:
                names = file.getNames(groups)
        return [name for i, name in enumerate(sorted(names)) if index is None or i in index]
    
    def getGroup(self):
        if self.is_directory:
            groups = set()
            for root, _, files in os.walk(self.filename):
                for file in files:
                    path = os.path.relpath(os.path.join(root, file.split(".")[0]), self.filename)
                    parts = path.split("/")
                    if len(parts) >= 2:
                        del parts[-2]
                    groups.add("/".join(parts))
        else:
            with Dataset.File(self.filename, True, self.format) as file:
                groups = file.getGroup()
        return list(groups)
    
    def getInfos(self, groups: str, name: str) -> tuple[list[int], Attribute]:
        if self.is_directory:
            for subDirectory in self._getSubDirectories(groups):
                group = groups.split("/")[-1]
                if os.path.exists("{}{}{}{}".format(self.filename, subDirectory, name, ".h5" if self.format == "h5" else "")):
                    with Dataset.File("{}{}{}".format(self.filename, subDirectory, name), True, self.format) as file:
                        result = file.getInfos("", group)      
        else:
            with Dataset.File(self.filename, True, self.format) as file:
                result = file.getInfos(groups, name)
        return result


import torch
import numpy as np
from abc import abstractmethod
from konfai.utils.config import config
from functools import partial

class Scheduler():

    def __init__(self, start_value: float) -> None:
        self.baseValue = float(start_value)
        self.it = 0
        
    def step(self, it: int):
        self.it = it

    @abstractmethod
    def get_value(self) -> float:
        pass

class Constant(Scheduler):

    def __init__(self, value: float = 1):
        super().__init__(value)

    def get_value(self) -> float:
        return self.baseValue

class CosineAnnealing(Scheduler):
    
    def __init__(self, start_value: float = 1, eta_min: float = 0.00001, T_max: int = 100):
        super().__init__(start_value)
        self.eta_min = eta_min
        self.T_max = T_max

    def get_value(self):
        return self.eta_min + (self.baseValue - self.eta_min) *(1 + np.cos(self.it * torch.pi / self.T_max)) / 2

class Warmup(torch.optim.lr_scheduler.LambdaLR):
    
    def warmup(warmup_steps: int, step: int) -> float:
        return min(1.0, (step+1) / (warmup_steps+1))

    def __init__(self, optimizer: torch.optim.Optimizer, warmup_steps: int = 10, last_epoch=-1, verbose="deprecated"):
        super().__init__(optimizer, partial(Warmup.warmup, warmup_steps), last_epoch, verbose)



import pytest
from pathlib import Path
from konfai.utils.config import config

# ---- Classe enfant pour les objets imbriqués ----
class SubObject:

    @config("SubObject")
    def __init__(self, z: float = 0.5):
        self.z = z

# ---- Classe principale testant tous les types pris en charge ----
class DummyAllTypes:
    @config("MainTest")
    def __init__(
        self,
        a: int = 1,
        b: str = "hello",
        c: bool = True,
        d: float = 3.14,
        e: list[int] = [1, 2, 3],
        f: list[str] = ["a", "b"],
        g: list[bool] = [True, False],
        h: list[float] = [1.0, 2.0],
        i: dict[str, int] = {"x": 1},
        j: SubObject = None,
        l: dict[str, SubObject] = {"A" : SubObject(), "B" : SubObject()}
    ):
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.e = e
        self.f = f
        self.g = g
        self.h = h
        self.i = i
        self.j = j
        self.l = l

@pytest.fixture
def dummy_config_file(tmp_path: Path):
    yaml_path = tmp_path / "config.yml"
    yaml_content = """
MainTest:
  a: 42
  b: test
  c: false
  d: 2.718
  e: [10, 20]
  f: [x, y]
  g: [true, false, true]
  h: [0.1, 0.2]
  i: {"foo": 9}
  j:
    z: 9.99
  l:
    key1:
      z: 7.7
    key2:
      z: 8.8
"""
    yaml_path.write_text(yaml_content)
    return yaml_path

def test_config_instantiation(monkeypatch):
    dummy_config_file = "./tests/dummy_data/dummy_config.yml"
    monkeypatch.setenv("KONFAI_CONFIG_FILE", str(dummy_config_file))
    monkeypatch.setenv("KONFAI_CONFIG_MODE", "Done")

    obj = DummyAllTypes(config=str(dummy_config_file))

    assert obj.a == 42
    assert obj.b == "test"
    assert obj.c is False
    assert abs(obj.d - 2.718) < 1e-6
    assert obj.e == [10, 20]
    assert obj.f == ["x", "y"]
    assert obj.g == [True, False, True]
    assert obj.h == [0.1, 0.2]
    assert obj.i == {"foo": 9}
    assert isinstance(obj.j, SubObject)
    assert obj.j.z == 9.99
    assert isinstance(obj.l, dict)
    assert set(obj.l.keys()) == {"key1", "key2"}
    assert obj.l["key1"].z == 7.7
    assert obj.l["key2"].z == 8.8
  
def test_config_create(monkeypatch):
    dummy_config_file = "./tests/dummy_data/dummy_default_config.yml"
    monkeypatch.setenv("KONFAI_CONFIG_FILE", str(dummy_config_file))
    monkeypatch.setenv("KONFAI_CONFIG_MODE", "Done")

    obj = DummyAllTypes(config=str(dummy_config_file))

    assert obj.a == 1
    assert obj.b == "hello"
    assert obj.c is True
    assert abs(obj.d - 3.14) < 1e-6
    assert obj.e == [1,2,3]
    assert obj.f == ["a", "b"]
    assert obj.g == [True, False]
    assert obj.h == [1, 2]
    assert obj.i == {}
    assert isinstance(obj.j, SubObject)
    assert obj.j.z == 0.5
    assert isinstance(obj.l, dict)
    assert set(obj.l.keys()) == {"A", "B"}
    assert obj.l["A"].z == 0.5
    assert obj.l["B"].z == 0.5


import os
from torch.utils.data import DataLoader
import torch
import tqdm
import numpy as np
import json
import shutil
import builtins
import importlib
from konfai import EVALUATIONS_DIRECTORY, PREDICTIONS_DIRECTORY, KONFAI_ROOT, CONFIG_FILE
from konfai.utils.config import config
from konfai.utils.utils import _getModule, DistributedObject, synchronize_data, EvaluatorError
from konfai.data.data_manager import DataMetric

class CriterionsAttr():

    @config()
    def __init__(self) -> None:
        pass  

class CriterionsLoader():

    @config()
    def __init__(self, criterionsLoader: dict[str, CriterionsAttr] = {"default:torch_nn_CrossEntropyLoss:Dice:NCC": CriterionsAttr()}) -> None:
        self.criterionsLoader = criterionsLoader

    def getCriterions(self, output_group : str, target_group : str) -> dict[torch.nn.Module, CriterionsAttr]:
        criterions = {}
        for module_classpath, criterionsAttr in self.criterionsLoader.items():
            module, name = _getModule(module_classpath, "konfai.metric.measure")
            criterions[config("{}.metrics.{}.targetsCriterions.{}.criterionsLoader.{}".format(KONFAI_ROOT(), output_group, target_group, module_classpath))(getattr(importlib.import_module(module), name))(config = None)] = criterionsAttr
        return criterions

class TargetCriterionsLoader():

    @config()
    def __init__(self, targetsCriterions : dict[str, CriterionsLoader] = {"default" : CriterionsLoader()}) -> None:
        self.targetsCriterions = targetsCriterions
        
    def getTargetsCriterions(self, output_group : str) -> dict[str, dict[torch.nn.Module, float]]:
        targetsCriterions = {}
        for target_group, criterionsLoader in self.targetsCriterions.items():
            targetsCriterions[target_group] = criterionsLoader.getCriterions(output_group, target_group)
        return targetsCriterions

class Statistics():

    def __init__(self, filename: str) -> None:
        self.measures: dict[str, dict[str, float]] = {}
        self.filename = filename

    def add(self, values: dict[str, float], name_dataset: str) -> None:
        for name, value in values.items():
            if name_dataset not in self.measures:
                self.measures[name_dataset] = {}
            self.measures[name_dataset][name] = value
    
    @staticmethod   
    def getStatistic(values: list[float]) -> dict[str, float]:
        return {"max": np.max(values), "min": np.min(values), "std": np.std(values), "25pc": np.percentile(values, 25), "50pc": np.percentile(values, 50), "75pc": np.percentile(values, 75), "mean": np.mean(values), "count": len(values)}
    
    def write(self, outputs: list[dict[str, any]]) -> None:
        measures = {}
        for output in outputs:
            measures.update(output)
        result = {}
        result["case"] = {}
        for name, v in measures.items():
            for metric_name, value in v.items():
                if metric_name not in result["case"]:
                    result["case"][metric_name] = {}
                result["case"][metric_name][name] = value

        result["aggregates"] = {}
        tmp: dict[str, list[float]] = {}
        for name, v in measures.items():
            for metric_name, value in v.items():
                if metric_name not in tmp:
                    tmp[metric_name] = []
                tmp[metric_name].append(v[metric_name])
        for metric_name, values in tmp.items():
            result["aggregates"][metric_name] = Statistics.getStatistic(values)

        with open(self.filename, "w") as f:
            f.write(json.dumps(result, indent=4))
            
class Evaluator(DistributedObject):

    @config("Evaluator")
    def __init__(self, train_name: str = "default:TRAIN_01", metrics: dict[str, TargetCriterionsLoader] = {"default": TargetCriterionsLoader()}, dataset : DataMetric = DataMetric(),) -> None:
        if os.environ["KONFAI_CONFIG_MODE"] != "Done":
            exit(0)
        super().__init__(train_name)
        self.metric_path = EVALUATIONS_DIRECTORY()+self.name+"/"
        self.metricsLoader = metrics
        self.dataset = dataset
        self.metrics = {k: v.getTargetsCriterions(k) for k, v in self.metricsLoader.items()}
        self.statistics_train = Statistics(self.metric_path+"Metric_TRAIN.json")
        self.statistics_validation = Statistics(self.metric_path+"Metric_VALIDATION.json")

    def update(self, data_dict: dict[str, tuple[torch.Tensor, str]], statistics : Statistics) -> dict[str, float]:
        result = {}
        for output_group in self.metrics:
            for target_group in self.metrics[output_group]:
                targets = [data_dict[group][0].to(0) if torch.cuda.is_available() else data_dict[group][0] for group in target_group.split(";") if group in data_dict]
                name = data_dict[output_group][1][0]
                for metric in self.metrics[output_group][target_group]:
                    result["{}:{}:{}".format(output_group, target_group, metric.__class__.__name__)] = metric(data_dict[output_group][0].to(0) if torch.cuda.is_available() else data_dict[output_group][0], *targets).item()
        statistics.add(result, name)
        return result
    
    def setup(self, world_size: int):
        if os.path.exists(self.metric_path):
            if os.environ["KONFAI_OVERWRITE"] != "True":
                accept = builtins.input("The metric {} already exists ! Do you want to overwrite it (yes,no) : ".format(self.name))
                if accept != "yes":
                    return
           
            if os.path.exists(self.metric_path):
                shutil.rmtree(self.metric_path) 

        if not os.path.exists(self.metric_path):
            os.makedirs(self.metric_path)
        metric_namefile_src = CONFIG_FILE().replace(".yml", "")
        shutil.copyfile(metric_namefile_src+".yml", "{}{}.yml".format(self.metric_path, metric_namefile_src))

        self.dataloader = self.dataset.getData(world_size)

        groupsDest = [group for groups in self.dataset.groups_src.values() for group in groups]

        missing_outputs = set(self.metrics.keys()) - set(groupsDest)
        if missing_outputs:
            raise EvaluatorError(
                f"The following metric output groups are missing from 'groupsDest': {sorted(missing_outputs)}. ",
                f"Available groups: {sorted(groupsDest)}"
            )

        target_groups = []
        for i in {target for targets in self.metrics.values() for target in targets}:
            for u in i.split(";"):
                target_groups.append(u)
        missing_targets = set(target_groups) - set(groupsDest)
        if missing_targets:
            raise EvaluatorError(
                f"The following metric target groups are missing from 'groupsDest': {sorted(missing_targets)}. ",
                f"Available groups: {sorted(groupsDest)}"
            )

    def run_process(self, world_size: int, global_rank: int, gpu: int, dataloaders: list[DataLoader]):
        description = lambda measure : "Metric TRAIN : {} ".format(" | ".join("{}: {:.2f}".format(k, v) for k, v in measure.items()) if measure is not None else "")
        with tqdm.tqdm(iterable = enumerate(dataloaders[0]), leave=True, desc = description(None), total=len(dataloaders[0]), ncols=0) as batch_iter:
            for _, data_dict in batch_iter:
                batch_iter.set_description(description(self.update({k: (v[0], v[4]) for k,v in data_dict.items()}, self.statistics_train)))
        outputs = synchronize_data(world_size, gpu, self.statistics_train.measures)
        if global_rank == 0:
            self.statistics_train.write(outputs)
        if len(dataloaders) == 2:
            description = lambda measure : "Metric VALIDATION : {} ".format(" | ".join("{}: {:.2f}".format(k, v) for k, v in measure.items()) if measure is not None else "")
            with tqdm.tqdm(iterable = enumerate(dataloaders[1]), leave=True, desc = description(None), total=len(dataloaders[1]), ncols=0) as batch_iter:
                for _, data_dict in batch_iter:
                    batch_iter.set_description(description(self.update({k: (v[0], v[4]) for k,v in data_dict.items()}, self.statistics_validation)))
            outputs = synchronize_data(world_size, gpu, self.statistics_validation.measures)
            if global_rank == 0:
                self.statistics_validation.write(outputs)


from functools import partial
import torch

from konfai.network import network, blocks
from konfai.utils.config import config
from konfai.data.patching import ModelPatch

class Discriminator(network.Network):

    class DiscriminatorNLayers(network.ModuleArgsDict):

        def __init__(self, channels: list[int], strides: list[int], dim: int) -> None:
            super().__init__()
            blockConfig = partial(blocks.BlockConfig, kernel_size=4, padding=1, bias=False, activation=partial(torch.nn.LeakyReLU, negative_slope = 0.2, inplace=True), normMode=blocks.NormMode.SYNCBATCH)
            for i, (in_channels, out_channels, stride) in enumerate(zip(channels, channels[1:], strides)):
                self.add_module("Layer_{}".format(i), blocks.ConvBlock(in_channels, out_channels, 1, blockConfig(stride=stride), dim))
    
    class DiscriminatorHead(network.ModuleArgsDict):

        def __init__(self, channels: int, dim: int) -> None:
            super().__init__()
            self.add_module("Conv", blocks.getTorchModule("Conv", dim)(in_channels=channels, out_channels=1, kernel_size=4, stride=1, padding=1))
            self.add_module("AdaptiveAvgPool", blocks.getTorchModule("AdaptiveAvgPool", dim)(tuple([1]*dim)))
            self.add_module("Flatten", torch.nn.Flatten(1))
    
    def __init__(self,
                    optimizer : network.OptimizerLoader = network.OptimizerLoader(),
                    schedulers: dict[str, network.LRSchedulersLoader] = {"default:ReduceLROnPlateau": network.LRSchedulersLoader(0)},
                    outputsCriterions: dict[str, network.TargetCriterionsLoader] = {"default" : network.TargetCriterionsLoader()},
                    nb_batch_per_step: int = 64,
                    dim : int = 3) -> None:
        super().__init__(in_channels = 1, optimizer = optimizer, schedulers = schedulers, outputsCriterions = outputsCriterions, dim=dim, nb_batch_per_step=nb_batch_per_step)
        channels = [1, 16, 32, 64, 64]
        strides = [2,2,2,1]
        self.add_module("Layers", Discriminator.DiscriminatorNLayers(channels, strides, dim))
        self.add_module("Head", Discriminator.DiscriminatorHead(channels[-1], dim))

class Generator(network.Network):

    class GeneratorStem(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module("ConvBlock", blocks.ConvBlock(in_channels, out_channels, nb_conv=1, blockConfig=blocks.BlockConfig(bias=False, activation="ReLU", normMode="SYNCBATCH"), dim=dim))

    class GeneratorHead(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module("ConvBlock", blocks.ConvBlock(in_channels, in_channels, nb_conv=1, blockConfig=blocks.BlockConfig(bias=False, activation="ReLU", normMode="SYNCBATCH"), dim=dim))
            self.add_module("Conv", blocks.getTorchModule("Conv", dim)(in_channels, out_channels, kernel_size=1, bias=False))
            self.add_module("Tanh", torch.nn.Tanh())

    class GeneratorDownSample(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module("ConvBlock", blocks.ConvBlock(in_channels, out_channels, nb_conv=1, blockConfig=blocks.BlockConfig(stride=2, bias=False, activation="ReLU", normMode="SYNCBATCH"), dim=dim))
    
    class GeneratorUpSample(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module("ConvBlock", blocks.ConvBlock(in_channels, out_channels, nb_conv=1, blockConfig=blocks.BlockConfig(bias=False, activation="ReLU", normMode="SYNCBATCH"), dim=dim))
            self.add_module("Upsample", torch.nn.Upsample(scale_factor=2, mode="bilinear" if dim < 3 else "trilinear"))
    
    class GeneratorEncoder(network.ModuleArgsDict):
        def __init__(self, channels: list[int], dim: int) -> None:
            super().__init__()
            for i, (in_channels, out_channels) in enumerate(zip(channels, channels[1:])):
                self.add_module("DownSample_{}".format(i), Generator.GeneratorDownSample(in_channels=in_channels, out_channels=out_channels, dim=dim))
    
    class GeneratorResnetBlock(network.ModuleArgsDict):

        def __init__(self, channels : int, dim : int):
            super().__init__()
            self.add_module("Conv_0", blocks.getTorchModule("Conv", dim)(channels, channels, kernel_size=3, padding=1, bias=False))
            self.add_module("Norm", torch.nn.LeakyReLU(0.2, inplace=True))
            self.add_module("Conv_1", blocks.getTorchModule("Conv", dim)(channels, channels, kernel_size=3, padding=1, bias=False))
            self.add_module("Residual", blocks.Add(), in_branch=[0,1])

    class GeneratorNResnetBlock(network.ModuleArgsDict):

        def __init__(self, channels: int, nb_conv: int, dim: int) -> None:
            super().__init__()
            for i in range(nb_conv):
                self.add_module("ResnetBlock_{}".format(i), Generator.GeneratorResnetBlock(channels=channels, dim=dim))

    class GeneratorDecoder(network.ModuleArgsDict):
        def __init__(self, channels: list[int], dim: int) -> None:
            super().__init__()
            for i, (in_channels, out_channels) in enumerate(zip(reversed(channels), reversed(channels[:-1]))):
                self.add_module("UpSample_{}".format(i), Generator.GeneratorUpSample(in_channels=in_channels, out_channels=out_channels, dim=dim))
    
    class GeneratorAutoEncoder(network.ModuleArgsDict):

        def __init__(self, ngf: int, dim: int) -> None:
            super().__init__()
            channels = [ngf, ngf*2]
            self.add_module("Encoder", Generator.GeneratorEncoder(channels, dim))
            self.add_module("NResBlock", Generator.GeneratorNResnetBlock(channels=channels[-1], nb_conv=6, dim=dim))
            self.add_module("Decoder", Generator.GeneratorDecoder(channels, dim))
        
    def __init__(self, 
                    optimizer : network.OptimizerLoader = network.OptimizerLoader(),
                    schedulers: dict[str, network.LRSchedulersLoader] = {"default:ReduceLROnPlateau": network.LRSchedulersLoader(0)},
                    patch : ModelPatch = ModelPatch(),
                    outputsCriterions: dict[str, network.TargetCriterionsLoader] = {"default" : network.TargetCriterionsLoader()},
                    nb_batch_per_step: int = 64,
                    dim : int = 3) -> None:
        super().__init__(optimizer=optimizer, in_channels=1, schedulers=schedulers, patch=patch, outputsCriterions=outputsCriterions, dim=dim, nb_batch_per_step=nb_batch_per_step)
        ngf=32
        self.add_module("Stem", Generator.GeneratorStem(1, ngf, dim))
        self.add_module("AutoEncoder", Generator.GeneratorAutoEncoder(ngf, dim))
        self.add_module("Head", Generator.GeneratorHead(in_channels=ngf, out_channels=1, dim=dim))

    def getName(self):
        return "Generator"

class Gan(network.Network):

    def __init__(self, generator : Generator = Generator(), discriminator : Discriminator = Discriminator()) -> None:
        super().__init__()
        self.add_module("Discriminator_B", discriminator, in_branch=[1], out_branch=[-1], requires_grad=True)
        self.add_module("Generator_A_to_B", generator, in_branch=[0], out_branch=["pB"])
        
        self.add_module("detach", blocks.Detach(), in_branch=["pB"], out_branch=["pB_detach"])
        self.add_module("Discriminator_pB_detach", discriminator, in_branch=["pB_detach"], out_branch=[-1])
          
        self.add_module("Discriminator_pB", discriminator, in_branch=["pB"], out_branch=[-1], requires_grad=False)




from enum import Enum
import importlib
from typing import Callable
import torch
from scipy.interpolate import interp1d
import numpy as np
import ast
from typing import Union

from konfai.utils.config import config
from konfai.network import network

class NormMode(Enum):
    NONE = 0,
    BATCH = 1
    INSTANCE = 2
    GROUP = 3
    LAYER = 4
    SYNCBATCH = 5
    INSTANCE_AFFINE = 6

def getNorm(normMode: Enum, channels : int, dim: int) -> Union[torch.nn.Module, None]:
    if normMode == NormMode.BATCH:
        return getTorchModule("BatchNorm", dim = dim)(channels, affine=True, track_running_stats=True)
    if normMode == NormMode.INSTANCE:
        return getTorchModule("InstanceNorm", dim = dim)(channels, affine=False, track_running_stats=False)
    if normMode == NormMode.INSTANCE_AFFINE:
        return getTorchModule("InstanceNorm", dim = dim)(channels, affine=True, track_running_stats=False)
    if normMode == NormMode.SYNCBATCH:
        return torch.nn.SyncBatchNorm(channels, affine=True, track_running_stats=True)
    if normMode == NormMode.GROUP:
        return torch.nn.GroupNorm(num_groups=32, num_channels=channels)
    if normMode == NormMode.LAYER:
        return torch.nn.GroupNorm(num_groups=1, num_channels=channels)
    return None

class UpSampleMode(Enum):
    CONV_TRANSPOSE = 0,
    UPSAMPLE = 1,

class DownSampleMode(Enum):
    MAXPOOL = 0,
    AVGPOOL = 1,
    CONV_STRIDE = 2

def getTorchModule(name_fonction : str, dim : Union[int, None] = None) -> torch.nn.Module:
    return getattr(importlib.import_module("torch.nn"), "{}".format(name_fonction) + ("{}d".format(dim) if dim is not None else ""))

class BlockConfig():

    @config("BlockConfig")
    def __init__(self, kernel_size : int = 3, stride : int = 1, padding : int = 1, bias = True, activation : Union[str, Callable[[], torch.nn.Module]] = "ReLU", normMode : Union[str, NormMode, Callable[[int], torch.nn.Module]] = "NONE") -> None:
        self.kernel_size = kernel_size
        self.bias = bias
        self.stride = stride
        self.padding = padding
        self.activation = activation
        if normMode is None:
            self.norm = None
        elif isinstance(normMode, str):
            self.norm = NormMode._member_map_[normMode]
        elif isinstance(normMode, NormMode):
            self.norm = normMode

    def getConv(self, in_channels : int, out_channels : int, dim : int) -> torch.nn.Conv3d:
        return getTorchModule("Conv", dim = dim)(in_channels = in_channels, out_channels = out_channels, kernel_size = self.kernel_size, stride = self.stride, padding = self.padding, bias=self.bias)
    
    def getNorm(self, channels : int, dim: int) -> torch.nn.Module:
        if self.norm is None:
            return None
        return getNorm(self.norm, channels, dim) if isinstance(self.norm, NormMode) else self.norm(channels)

    def getActivation(self) -> torch.nn.Module:
        if self.activation is None:
            return None
        if isinstance(self.activation, str):
            return getTorchModule(self.activation.split(";")[0])(*[ast.literal_eval(value) for value in self.activation.split(";")[1:]]) if self.activation != "None" else torch.nn.Identity()
        return self.activation()
    
class ConvBlock(network.ModuleArgsDict):
    
    def __init__(self, in_channels : int, out_channels : int, blockConfigs : list[BlockConfig], dim : int, alias : list[list[str]]=[[], [], []]) -> None:
        super().__init__()
        for i, blockConfig in enumerate(blockConfigs):
            self.add_module("Conv_{}".format(i), blockConfig.getConv(in_channels, out_channels, dim), alias=alias[0])
            norm = blockConfig.getNorm(out_channels, dim)
            if norm is not None:
                self.add_module("Norm_{}".format(i), norm, alias=alias[1])
            activation = blockConfig.getActivation()
            if activation is not None:
                self.add_module("Activation_{}".format(i), activation, alias=alias[2])
            in_channels = out_channels

class ResBlock(network.ModuleArgsDict):
    
    def __init__(self, in_channels : int, out_channels : int, blockConfigs : list[BlockConfig], dim : int, alias : list[list[str]]=[[], [], [], [], []]) -> None:
        super().__init__()
        for i, blockConfig in enumerate(blockConfigs):
            self.add_module("Conv_{}".format(i), blockConfig.getConv(in_channels, out_channels, dim), alias=alias[0])
            norm = blockConfig.getNorm(out_channels, dim)
            if norm is not None:
                self.add_module("Norm_{}".format(i), norm, alias=alias[1])
            activation = blockConfig.getActivation()
            if activation is not None:
                self.add_module("Activation_{}".format(i), activation, alias=alias[2])

            if in_channels != out_channels:
                self.add_module("Conv_skip", getTorchModule("Conv", dim)(in_channels, out_channels, 1, blockConfig.stride, bias=blockConfig.bias), alias=alias[3], in_branch=[1], out_branch=[1])    
                self.add_module("Norm_skip", blockConfig.getNorm(out_channels, dim), alias=alias[4], in_branch=[1], out_branch=[1])    
            in_channels = out_channels
            
        self.add_module("Add", Add(), in_branch=[0,1])
        self.add_module("Norm_{}".format(i+1), torch.nn.ReLU(inplace=True))
            
def downSample(in_channels: int, out_channels: int, downSampleMode: DownSampleMode, dim: int) -> torch.nn.Module:
    if downSampleMode == DownSampleMode.MAXPOOL:
        return getTorchModule("MaxPool", dim = dim)(2)
    if downSampleMode == DownSampleMode.AVGPOOL:
        return getTorchModule("AvgPool", dim = dim)(2)
    if downSampleMode == DownSampleMode.CONV_STRIDE:
        return getTorchModule("Conv", dim)(in_channels, out_channels, kernel_size=2, stride=2, padding=0)

def upSample(in_channels: int, out_channels: int, upSampleMode: UpSampleMode, dim: int, kernel_size: Union[int, list[int]] = 2, stride: Union[int, list[int]] = 2):
    if upSampleMode == UpSampleMode.CONV_TRANSPOSE:
        return getTorchModule("ConvTranspose", dim = dim)(in_channels = in_channels, out_channels = out_channels, kernel_size = kernel_size, stride = stride, padding = 0)
    else:
        if dim == 3:
            upsampleMethod = "trilinear" 
        if dim == 2:
            upsampleMethod = "bilinear"
        if dim == 1:
            upsampleMethod = "linear"
        return torch.nn.Upsample(scale_factor=2, mode=upsampleMethod.lower(), align_corners=False)

class Unsqueeze(torch.nn.Module):

    def __init__(self, dim: int = 0):
        super().__init__()
        self.dim = dim
    
    def forward(self, *input : torch.Tensor) -> torch.Tensor:
        return torch.unsqueeze(input, self.dim)
    
    def extra_repr(self):
        return "dim={}".format(self.dim)

class Permute(torch.nn.Module):

    def __init__(self, dims : list[int]):
        super().__init__()
        self.dims = dims

    def forward(self, input : torch.Tensor) -> torch.Tensor:
        return torch.permute(input, self.dims)
    
    def extra_repr(self):
        return "dims={}".format(self.dims)

class ToChannels(Permute):

    def __init__(self, dim):
        super().__init__([0, dim+1, *[i+1 for i in range(dim)]])
        
class ToFeatures(Permute):

    def __init__(self, dim):
        super().__init__([0, *[i+2 for i in range(dim)], 1])        

class Add(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, *input : torch.Tensor) -> torch.Tensor:
        return torch.sum(torch.stack(input), dim=0)
    
class Multiply(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()

    def forward(self, *input : torch.Tensor) -> torch.Tensor:
        return torch.mul(*input)

class Concat(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()

    def forward(self, *input : torch.Tensor) -> torch.Tensor:
        return torch.cat(input, dim=1)

class Print(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        print(input.shape)
        return input

class Write(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        import SimpleITK as sitk
        sitk.WriteImage(sitk.GetImageFromArray(input.clone()[0][0].cpu().numpy()), "./Data.mha")
        return input
    
class Exit(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        exit(0)
    
class Detach(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return input.detach()
    
class Negative(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()
    
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return -input

class GetShape(torch.nn.Module):

    def __init__(self) -> None:
        super().__init__()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return torch.tensor(input.shape)

class ArgMax(torch.nn.Module):

    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return torch.argmax(input, dim=self.dim).unsqueeze(self.dim)
    
class Select(torch.nn.Module):

    def __init__(self, slices: list[slice]) -> None:
        super().__init__()
        self.slices = tuple(slices)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        result = input[self.slices]
        for i, s in enumerate(range(len(result.shape))):
            if s == 1:
              result = result.squeeze(dim=i)  
        return result

class NormalNoise(torch.nn.Module):

    def __init__(self, dim: Union[int, None] = None) -> None:
        super().__init__()
        self.dim = dim
    
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.dim is not None:
            return torch.randn(self.dim).to(input.device)
        else:
            return torch.randn_like(input).to(input.device)
    
class Const(torch.nn.Module):

    def __init__(self, shape: list[int], std: float) -> None:
        super().__init__()
        self.noise = torch.nn.parameter.Parameter(torch.randn(shape)*std)
        
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return self.noise.to(input.device)

class HistogramNoise(torch.nn.Module):

    def __init__(self, n: int, sigma: float) -> None:
        super().__init__()
        self.x = np.linspace(0, 1, num=n, endpoint=True)
        self.sigma = sigma
    
    def forward(self, input: torch.Tensor) -> torch.Tensor:
        self.function = interp1d(self.x, self.x+np.random.normal(0, self.sigma, self.x.shape[0]), kind='cubic')
        result = torch.empty_like(input)

        for value in torch.unique(input):
            x = self.function(value.cpu())
            result[torch.where(input == value)] = torch.tensor(x, device=input.device).float()
        return result

class Subset(torch.nn.Module):
	def __init__(self, slices: list[slice]):
		super().__init__()
		self.slices = [slice(None, None), slice(None, None)] + slices

	def forward(self, tensor: torch.Tensor) -> torch.Tensor:
		return tensor[self.slices]

class View(torch.nn.Module):
	def __init__(self, size: list[int]):
		super().__init__()
		self.size = size

	def forward(self, tensor: torch.Tensor) -> torch.Tensor:
		return tensor.view(self.size)
        
class LatentDistribution(network.ModuleArgsDict):

    class LatentDistribution_Linear(torch.nn.Module):

        def __init__(self, shape: list[int], latentDim: int) -> None:
            super().__init__()
            self.linear = torch.nn.Linear(torch.prod(torch.tensor(shape)), latentDim)

        def forward(self, input: torch.Tensor) -> torch.Tensor:
            return torch.unsqueeze(self.linear(input), 1)
        
    class LatentDistribution_Decoder(torch.nn.Module):
        
        def __init__(self, shape: list[int], latentDim: int) -> None:
            super().__init__()
            self.linear = torch.nn.Linear(latentDim, torch.prod(torch.tensor(shape)))
            self.shape = shape

        def forward(self, input: torch.Tensor) -> torch.Tensor:
            return self.linear(input).view(-1, *[int(i) for i in self.shape])
    
    class LatentDistribution_Z(torch.nn.Module):

        def __init__(self) -> None:
            super().__init__()

        def forward(self, mu: torch.Tensor, log_std: torch.Tensor) -> torch.Tensor:
            return torch.exp(log_std/2)*torch.rand_like(mu)+mu
    
    def __init__(self, shape: list[int], latentDim: int) -> None:
        super().__init__()        
        self.add_module("Flatten", torch.nn.Flatten(1))
        self.add_module("mu", LatentDistribution.LatentDistribution_Linear(shape, latentDim), out_branch = [1])
        self.add_module("log_std", LatentDistribution.LatentDistribution_Linear(shape, latentDim), out_branch = [2])

        self.add_module("z", LatentDistribution.LatentDistribution_Z(), in_branch=[1,2], out_branch=[3])
        self.add_module("Concat", Concat(), in_branch=[1,2,3])
        self.add_module("DecoderInput", LatentDistribution.LatentDistribution_Decoder(shape, latentDim), in_branch=[3])

class Attention(network.ModuleArgsDict):

    def __init__(self, F_g : int, F_l : int, F_int : int, dim : int):
        super().__init__()
        self.add_module("W_x", getTorchModule("Conv", dim = dim)(in_channels = F_l, out_channels = F_int, kernel_size=1, stride=2, padding=0), in_branch=[0], out_branch=[0])
        self.add_module("W_g", getTorchModule("Conv", dim = dim)(in_channels = F_g, out_channels = F_int, kernel_size=1, stride=1, padding=0), in_branch=[1], out_branch=[1])
        self.add_module("Add", Add(), in_branch=[0,1])
        self.add_module("ReLU", torch.nn.ReLU(inplace=True))
        self.add_module("Conv", getTorchModule("Conv", dim = dim)(in_channels = F_int, out_channels = 1, kernel_size=1,stride=1, padding=0))
        self.add_module("Sigmoid", torch.nn.Sigmoid())
        self.add_module("Upsample", torch.nn.Upsample(scale_factor=2))
        self.add_module("Multiply", Multiply(), in_branch=[2,0])


import torch

from konfai.network import network, blocks
from konfai.utils.config import config

class VAE(network.Network):

    class AutoEncoderBlock(network.ModuleArgsDict):

        def __init__(self, channels: list[int], nb_conv_per_stage: int, blockConfig: blocks.BlockConfig, downSampleMode: blocks.DownSampleMode, upSampleMode: blocks.UpSampleMode, dim: int, block: type, i : int = 0) -> None:
            super().__init__()
            if i > 0:
                self.add_module(downSampleMode.name, blocks.downSample(in_channels=channels[0], out_channels=channels[1], downSampleMode=downSampleMode, dim=dim))
            self.add_module("DownBlock", block(in_channels=channels[1 if downSampleMode == blocks.DownSampleMode.CONV_STRIDE and i > 0 else 0], out_channels=channels[1], blockConfigs=[blockConfig]*nb_conv_per_stage, dim=dim))
            if len(channels) > 2:
                self.add_module("AutoEncoder_{}".format(i+1), VAE.AutoEncoderBlock(channels[1:], nb_conv_per_stage, blockConfig, downSampleMode, upSampleMode, dim, block, i+1))
                self.add_module("UpBlock", block(in_channels=channels[2] if upSampleMode != blocks.UpSampleMode.CONV_TRANSPOSE else channels[1], out_channels=channels[1], blockConfigs=[blockConfig]*nb_conv_per_stage, dim=dim))
            if i > 0:
               self.add_module(upSampleMode.name, blocks.upSample(in_channels=channels[1], out_channels=channels[0], upSampleMode=upSampleMode, dim=dim))
    
    class VAE_Head(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module("Conv", blocks.getTorchModule("Conv", dim)(in_channels = in_channels, out_channels = out_channels, kernel_size = 3, stride = 1, padding = 1))
            self.add_module("Tanh", torch.nn.Tanh()) 

    def __init__(self,
                    optimizer: network.OptimizerLoader = network.OptimizerLoader(),
                    schedulers: network.LRSchedulersLoader = network.LRSchedulersLoader(),
                    outputsCriterions: dict[str, network.TargetCriterionsLoader] = {"default" : network.TargetCriterionsLoader()},
                    dim : int = 3,
                    channels: list[int]=[1, 64, 128, 256, 512, 1024],
                    blockConfig: blocks.BlockConfig = blocks.BlockConfig(),
                    nb_conv_per_stage: int = 2,
                    downSampleMode: str = "MAXPOOL",
                    upSampleMode: str = "CONV_TRANSPOSE",
                    blockType: str = "Conv") -> None:
        
        super().__init__(in_channels = channels[0], init_type="normal", optimizer = optimizer, schedulers = schedulers, outputsCriterions = outputsCriterions, dim=dim, nb_batch_per_step=1)
        self.add_module("AutoEncoder_0", VAE.AutoEncoderBlock(channels, nb_conv_per_stage, blockConfig, downSampleMode=blocks.DownSampleMode._member_map_[downSampleMode], upSampleMode=blocks.UpSampleMode._member_map_[upSampleMode], dim=dim, block = blocks.ConvBlock if blockType == "Conv" else blocks.ResBlock))
        self.add_module("Head", VAE.VAE_Head(channels[1], channels[0], dim))


class LinearVAE(network.Network):

    class LinearVAE_DenseLayer(network.ModuleArgsDict):

        def __init__(self, in_features: int, out_features: int) -> None:
            super().__init__()
            self.add_module("Linear", torch.nn.Linear(in_features, out_features))
            #self.add_module("Norm", torch.nn.BatchNorm1d(out_features))
            self.add_module("Activation", torch.nn.LeakyReLU())

    class LinearVAE_Head(network.ModuleArgsDict):

        def __init__(self, in_features: int, out_features: int) -> None:
            super().__init__()
            self.add_module("Linear", torch.nn.Linear(in_features, out_features))
            self.add_module("Tanh", torch.nn.Tanh())

    def __init__(self,
                    optimizer: network.OptimizerLoader = network.OptimizerLoader(),
                    schedulers: network.LRSchedulersLoader = network.LRSchedulersLoader(),
                    outputsCriterions: dict[str, network.TargetCriterionsLoader] = {"default" : network.TargetCriterionsLoader()},) -> None:
        super().__init__(in_channels = 1, init_type="normal", optimizer = optimizer, schedulers = schedulers, outputsCriterions = outputsCriterions, dim=1, nb_batch_per_step=1)
        self.add_module("DenseLayer_0", LinearVAE.LinearVAE_DenseLayer(23343, 5))
        #self.add_module("Head", LinearVAE.DenseLayer(100, 28590))
        self.add_module("Head", LinearVAE.LinearVAE_Head(5, 23343))
        #self.add_module("DenseLayer_5", LinearVAE.DenseLayer(5000, 28590))


import importlib
import torch

from konfai.network import network, blocks
from konfai.utils.config import config
from konfai.data.patching import ModelPatch    

class MappingNetwork(network.ModuleArgsDict):
    def __init__(self, z_dim: int, c_dim: int, w_dim: int, num_layers: int, embed_features: int, layer_features: int):
        super().__init__()
        
        self.add_module("Concat_1", blocks.Concat(), in_branch=[0,1])
        
        features = [z_dim + embed_features if c_dim > 0 else 0] + [layer_features] * (num_layers - 1) + [w_dim]    
        if c_dim > 0:
            self.add_module("Linear", torch.nn.Linear(c_dim, embed_features), out_branch=["Embed"])

        self.add_module("Noise", blocks.NormalNoise(z_dim), in_branch=["Embed"])
        if c_dim > 0:
            self.add_module("Concat", blocks.Concat(), in_branch=[0,"Embed"])
        
        for i, (in_features, out_features) in enumerate(zip(features, features[1:])):
            self.add_module("Linear_{}".format(i), torch.nn.Linear(in_features, out_features))
    
class ModulatedConv(torch.nn.Module):

    class _ModulatedConv(torch.nn.Module):

        def __init__(self, w_dim: int, conv: torch.nn.modules.conv._ConvNd, dim: int) -> None:
            super().__init__()
            self.affine = torch.nn.Linear(w_dim, conv.in_channels)
            self.isConv = True
            self.in_channels = conv.in_channels
            self.out_channels = conv.out_channels
            self.padding = conv.padding
            self.stride = conv.stride
            if isinstance(conv, torch.nn.modules.conv._ConvTransposeNd):
                self.weight = torch.nn.parameter.Parameter(torch.randn((conv.in_channels, conv.out_channels, *conv.kernel_size)))
                self.isConv = False
            else:
                self.weight = torch.nn.parameter.Parameter(torch.randn((conv.out_channels, conv.in_channels, *conv.kernel_size)))
            conv.forward = self.forward
            self.styles = None
            self.dim = dim

        def setStyle(self, styles: torch.Tensor) -> None:
            self.styles = styles

        def forward(self, input: torch.Tensor) -> torch.Tensor:
            b = input.shape[0]
            self.affine.to(input.device)
            styles = self.affine(self.styles)
            w1 = styles.reshape(b, -1, 1, *[1 for _ in range(self.dim)]) if not self.isConv else styles.reshape(b, 1, -1, *[1 for _ in range(self.dim)])
            w2 = self.weight.unsqueeze(0).to(input.device)
            weights = w2 * (w1 + 1)

            d = torch.rsqrt((weights ** 2).sum(dim=tuple([i+2 for i in range(len(weights.shape)-2)]), keepdim=True) + 1e-8)
            weights = weights * d
            
            input = input.reshape(1, -1, *input.shape[2:])

            _, _, *ws = weights.shape
            if not self.isConv:
                out = getattr(importlib.import_module("torch.nn.functional"), "conv_transpose{}d".format(self.dim))(input, weights.reshape(b * self.in_channels, *ws), stride=self.stride, padding=self.padding, groups=b)
            else:
                out = getattr(importlib.import_module("torch.nn.functional"), "conv{}d".format(self.dim))(input, weights.reshape(b * self.out_channels, *ws), padding=self.padding, groups=b, stride=self.stride)
            
            out = out.reshape(-1, self.out_channels, *out.shape[2:])
            return out
    
    def __init__(self, w_dim: int, module: torch.nn.Module) -> None:
        super().__init__()
        self.w_dim = w_dim
        self.module = module
        self.convs = torch.nn.ModuleList() 
        self.module.apply(self.apply)
        
    def forward(self, input: torch.Tensor, styles: torch.Tensor) -> torch.Tensor:
        for conv in self.convs:
            conv.setStyle(styles.clone())
        return self.module(input)

    def apply(self, module: torch.nn.Module):
        if isinstance(module, torch.nn.modules.conv._ConvNd):
            delattr(module, "weight")
            module.bias = None
            
            str_dim = module.__class__.__name__[-2:]
            dim = 1
            if str_dim == "2d":
                dim = 2
            elif str_dim == "3d":
                dim = 3
            self.convs.append(ModulatedConv._ModulatedConv(self.w_dim, module, dim=dim))
        
class UNetBlock(network.ModuleArgsDict):

    def __init__(self, w_dim: int, channels: list[int], nb_conv_per_stage: int, blockConfig: blocks.BlockConfig, downSampleMode: blocks.DownSampleMode, upSampleMode: blocks.UpSampleMode, attention : bool, dim: int, i : int = 0) -> None:
        super().__init__()
        if i > 0:
            self.add_module(downSampleMode.name, blocks.downSample(in_channels=channels[0], out_channels=channels[1], downSampleMode=downSampleMode, dim=dim))
        self.add_module("DownConvBlock", blocks.ConvBlock(in_channels=channels[1 if downSampleMode == blocks.DownSampleMode.CONV_STRIDE and i > 0 else 0], out_channels=channels[1], nb_conv=nb_conv_per_stage, blockConfig=blockConfig, dim=dim))
        if len(channels) > 2:
            self.add_module("UNetBlock_{}".format(i+1), UNetBlock(w_dim, channels[1:], nb_conv_per_stage, blockConfig, downSampleMode, upSampleMode, attention, dim, i+1), in_branch=[0,1])
            self.add_module("UpConvBlock", ModulatedConv(w_dim, blocks.ConvBlock((channels[1]+channels[2]) if upSampleMode != blocks.UpSampleMode.CONV_TRANSPOSE else channels[1]*2, out_channels=channels[1], nb_conv=nb_conv_per_stage, blockConfig=blockConfig, dim=dim)), in_branch=[0,1])
        if i > 0:
            if attention:
                self.add_module("Attention", blocks.Attention(F_g=channels[1], F_l=channels[0], F_int=channels[0], dim=dim), in_branch=["Skip", 0], out_branch=["Skip"])
            self.add_module(upSampleMode.name, ModulatedConv(w_dim, blocks.upSample(in_channels=channels[1], out_channels=channels[0], upSampleMode=upSampleMode, dim=dim)), in_branch=[0, 1])
            self.add_module("SkipConnection", blocks.Concat(), in_branch=[0, "Skip"])

class Generator(network.Network):

    class GeneratorHead(network.ModuleArgsDict):

        def __init__(self, in_channels: int, out_channels: int, dim: int) -> None:
            super().__init__()
            self.add_module("Conv", blocks.getTorchModule("Conv", dim)(in_channels = in_channels, out_channels = out_channels, kernel_size = 3, stride = 1, padding = 1))
            self.add_module("Tanh", torch.nn.Tanh())

    def __init__(self, 
                    optimizer : network.OptimizerLoader = network.OptimizerLoader(),
                    schedulers: dict[str, network.LRSchedulersLoader] = {"default:ReduceLROnPlateau": network.LRSchedulersLoader(0)},
                    patch : ModelPatch = ModelPatch(),
                    outputsCriterions: dict[str, network.TargetCriterionsLoader] = {"default" : network.TargetCriterionsLoader()},
                    channels: list[int]=[1, 64, 128, 256, 512, 1024],
                    nb_batch_per_step: int = 64,
                    z_dim: int = 512,
                    c_dim: int = 1,
                    w_dim: int = 512,
                    dim : int = 3) -> None:
        super().__init__(optimizer=optimizer, in_channels=channels[0], schedulers=schedulers, patch=patch, outputsCriterions=outputsCriterions, dim=dim, nb_batch_per_step=nb_batch_per_step)

        self.add_module("MappingNetwork", MappingNetwork(z_dim=z_dim, c_dim=c_dim, w_dim=w_dim, num_layers=8, embed_features=w_dim, layer_features=w_dim), in_branch=[1,2], out_branch=["Style"])
        nb_conv_per_stage = 2
        blockConfig = blocks.BlockConfig(kernel_size=3, stride=1, padding=1, bias=True, activation="ReLU", normMode="INSTANCE") 
        self.add_module("UNetBlock_0", UNetBlock(w_dim, channels, nb_conv_per_stage, blockConfig, downSampleMode=blocks.DownSampleMode.MAXPOOL, upSampleMode=blocks.UpSampleMode.CONV_TRANSPOSE, attention=False, dim=dim), in_branch=[0, "Style"])
        self.add_module("Head", Generator.GeneratorHead(in_channels=channels[1], out_channels=1, dim=dim))